import pandas as pd 
import os 
from .io_utils import read_hdf
import numpy as np
__refdata__ = os.environ.get('picaso_refdata')
import pickle as pk
import astropy.constants as c
import scipy.interpolate as sci

def get_cld_input_grid(filename_or_grid):
	"""
	The albedo code relies on the cloud code input, which is traditionally on a 196 wavelength grid. 
	This method is to retrieve that grid. This file should be kept in the package reference data. Alternatively, 
	it might be useful to add it directly to the input.cld file. 

	Parameters
	----------
	filename_or_grid : str or ndarray
		filename of the input grid OR a numpy array of wavenumber corresponding to cloud 
		input

	Returns 
	-------
	array 
		array of wave numbers in increasing order 
	"""
	if filename_or_grid == 'wave_EGP.dat':
		grid = pd.read_csv(os.path.join(__refdata__, 'opacities',filename_or_grid), delim_whitespace=True)
		grid = grid.sort_values('wavenumber')['wavenumber'].values
	elif isinstance(filename_or_grid, np.ndarray):
		grid = np.sort(filename_or_grid)
	elif (isinstance(filename_or_grid, str) & (filename_or_grid != 'wave_EGP.dat') & 
		os.path.exists(filename_or_grid)):	
		grid = pd.read_csv(os.path.join(filename_or_grid), delim_whitespace=True)
		if 'wavenumber' in grid.keys():
			grid = grid.sort_values('wavenumber')['wavenumber'].values
		else: 
			raise Exception('Please make sure there is a column named "wavenumber" in your cloud wavegrid file')
	else:
		raise Exception("Please enter valid cloud wavegrid filepath, or numpy array. Or use default in reference file.")

	return grid

def regrid(matrix, old_wno, new_wno):
	"""
	This takes in a matrix that is (number of something versus number of wavelength points) and regrids 
	to a new wave grid. 

	Parameters
	----------
	matrix : ndarray
		matrix that is (number of something versus number of wavelength points)
	old_wno : array
		array that represents the old wavelength grid of the matrix 
	new_wno : array
		array that represents the desired wavelength grid 

	Returns
	-------
	matrix 
		matrix that has been reinterpolated to new wave number grid 
	"""
	new = np.zeros((matrix.shape[0],len(new_wno)))
	for i in range(matrix.shape[0]): 
		new[i, :] = np.interp(new_wno, old_wno, matrix[i,:])
		#f = sci.interp1d(old_wno, matrix[i,:],kind='cubic')
		#new[i, :] = f(new_wno)
	return new


from .atmsetup import ATMSETUP
from .fluxes import get_reflected_1d, get_reflected_3d , get_thermal_1d
from .wavelength import get_cld_input_grid
import numpy as np
import pandas as pd
from .optics import RetrieveOpacities,compute_opacity
from scipy.interpolate import RegularGridInterpolator
import scipy as sp
from scipy import special
import os
import pickle as pk
from .disco import get_angles, compute_disco, compress_disco, compress_thermal
import copy
import json
import pysynphot as psyn
import astropy.units as u
import astropy.constants as c
__refdata__ = os.environ.get('picaso_refdata')

def picaso(bundle,opacityclass, dimension = '1d',calculation='reflected', full_output=False, plot_opacity= False):
	"""
	Currently top level program to run albedo code 

	Parameters 
	----------
	bundle : dict 
		This input dict is built by loading the input = `justdoit.load_inputs()` 
	dimension : str 
		(Optional) Dimensions of the calculation. Default = '1d'. But '3d' is also accepted. 
		In order to run '3d' calculations, user must build 3d input (see tutorials)
	full_output : bool 
		(Optional) Default = False. Returns atmosphere class, which enables several 
		plotting capabilities. 
	plot_opacity : bool 
		(Optional) Default = False, Creates pop up of the weighted opacity

	Return
	------
	Wavenumber, albedo if full_output=False 
	Wavenumber, albedo, atmosphere if full_output = True 
	"""
	inputs = bundle.inputs

	wno = opacityclass.wno
	nwno = opacityclass.nwno

	#check to see if we are running in test mode
	test_mode = inputs['test_mode']

	#set approx numbers options (to be used in numba compiled functions)
	single_phase = inputs['approx']['single_phase']
	multi_phase = inputs['approx']['multi_phase']
	raman_approx =inputs['approx']['raman']

	#phase angle 
	phase_angle = inputs['phase_angle']

	#parameters needed for the two term hg phase function. 
	#Defaults are set in config.json
	f = inputs['approx']['TTHG_params']['fraction']
	frac_a = f[0]
	frac_b = f[1]
	frac_c = f[2]
	constant_back = inputs['approx']['TTHG_params']['constant_back']
	constant_forward = inputs['approx']['TTHG_params']['constant_forward']

	#get geometry
	ng = inputs['disco']['num_gangle']
	nt = inputs['disco']['num_tangle']

	gangle,gweight,tangle,tweight = get_angles(ng, nt) 
	#planet disk is divided into gaussian and chebyshev angles and weights for perfoming the 
	#intensity as a function of planetary pahse angle 
	ubar0, ubar1, cos_theta,lat,lon = compute_disco(ng, nt, gangle, tangle, phase_angle)

	inputs['disco']['latitude'] = lat
	inputs['disco']['longitude'] = lon

	#set star 
	radius_star = inputs['star']['radius']
	F0PI = np.zeros(nwno) + 1.0 

	#define approximinations 
	delta_eddington = inputs['approx']['delta_eddington']

	#begin atm setup
	atm = ATMSETUP(inputs)

	################ From here on out is everything that would go through retrieval or 3d input##############
	atm.planet.gravity = inputs['planet']['gravity']
	atm.planet.radius = inputs['planet']['radius']

	if dimension == '1d':
		atm.get_profile()
	elif dimension == '3d':
		atm.get_profile_3d()

	#now can get these 
	atm.get_mmw()
	atm.get_density()
	atm.get_column_density()
	#get needed continuum molecules 
	atm.get_needed_continuum()

	#get cloud properties, if there are any and put it on current grid 
	atm.get_clouds(wno)

	#determine surface reflectivity as function of wavelength (set to zero here)
	#TODO: Should be an input
	atm.get_surf_reflect(nwno) 

	#Make sure that all molecules are in opacityclass. If not, remove them and add warning
	no_opacities = [i for i in atm.molecules if i not in opacityclass.molecules]
	atm.add_warnings('No computed opacities for: '+','.join(no_opacities))
	atm.molecules = np.array([ x for x in atm.molecules if x not in no_opacities ])

	

	if dimension == '1d':
		#lastly grab needed opacities for the problem
		opacityclass.get_opacities(atm)
		#only need to get opacities for one pt profile

		#There are two sets of dtau,tau,w0,g in the event that the user chooses to use delta-eddington
		#We use HG function for single scattering which gets the forward scattering/back scattering peaks 
		#well. We only really want to use delta-edd for multi scattering legendre polynomials. 
		DTAU, TAU, W0, COSB,ftau_cld, ftau_ray,GCOS2, DTAU_OG, TAU_OG, W0_OG, COSB_OG= compute_opacity(
			atm, opacityclass,delta_eddington=delta_eddington,test_mode=test_mode,raman=raman_approx,
			full_output=full_output, plot_opacity=plot_opacity)

		if  'reflected' in calculation:
			#use toon method (and tridiagonal matrix solver) to get net cumulative fluxes 
			xint_at_top  = get_reflected_1d(atm.c.nlevel, wno,nwno,ng,nt,
													DTAU, TAU, W0, COSB,GCOS2,ftau_cld,ftau_ray,
													DTAU_OG, TAU_OG, W0_OG, COSB_OG ,
													atm.surf_reflect, ubar0,ubar1,cos_theta, F0PI,
													single_phase,multi_phase,
													frac_a,frac_b,frac_c,constant_back,constant_forward)

		if 'thermal' in calculation:
			#use toon method (and tridiagonal matrix solver) to get net cumulative fluxes 
			ng_therm = 8 #this is always 8 because in 1d we don't care where the flux is coming from 
			nt_therm = 2 #ditto
			gangle_therm,gweight_therm,tangle_therm,tweight_therm = get_angles(ng_therm, nt_therm) 
			j, ubar1_therm, j,j,j = compute_disco(ng_therm, nt_therm, gangle_therm, tangle_therm, phase_angle)
			flux_at_top  = get_thermal_1d(atm.c.nlevel, wno,nwno,ng_therm,nt_therm,atm.level['temperature'],
													DTAU_OG, W0_OG, COSB_OG, atm.level['pressure'],ubar1_therm)
			
	elif dimension == '3d':

		#setup zero array to fill with opacities
		TAU_3d = np.zeros((atm.c.nlevel, nwno, ng, nt))
		DTAU_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		W0_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		COSB_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		GCOS2_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		FTAU_CLD_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		FTAU_RAY_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		#these are the unchanged values from delta-eddington
		TAU_OG_3d = np.zeros((atm.c.nlevel, nwno, ng, nt))
		DTAU_OG_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		W0_OG_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))
		COSB_OG_3d = np.zeros((atm.c.nlayer, nwno, ng, nt))

		#get opacities at each facet
		for g in range(ng):
			for t in range(nt): 

				#edit atm class to only have subsection of 3d stuff 
				atm_1d = copy.deepcopy(atm)

				#diesct just a subsection to get the opacity 
				atm_1d.disect(g,t)

				opacityclass.get_opacities(atm_1d)

				dtau, tau, w0, cosb,ftau_cld, ftau_ray, gcos2, DTAU_OG, TAU_OG, W0_OG, COSB_OG = compute_opacity(
					atm_1d, opacityclass,delta_eddington=delta_eddington,test_mode=test_mode,raman=raman_approx)
				DTAU_3d[:,:,g,t] = dtau
				TAU_3d[:,:,g,t] = tau
				W0_3d[:,:,g,t] = w0 
				COSB_3d[:,:,g,t] = cosb
				GCOS2_3d[:,:,g,t]= gcos2 
				FTAU_CLD_3d[:,:,g,t]= ftau_cld
				FTAU_RAY_3d[:,:,g,t]= ftau_ray
				#these are the unchanged values from delta-eddington
				TAU_OG_3d[:,:,g,t] = TAU_OG
				DTAU_OG_3d[:,:,g,t] = DTAU_OG
				W0_OG_3d[:,:,g,t] = W0_OG
				COSB_OG_3d[:,:,g,t] = COSB_OG
		#use toon method (and tridiagonal matrix solver) to get net cumulative fluxes 
		xint_at_top  = get_reflected_3d(atm.c.nlevel, wno,nwno,ng,nt,
											DTAU_3d, TAU_3d, W0_3d, COSB_3d,GCOS2_3d, FTAU_CLD_3d,FTAU_RAY_3d,
											DTAU_OG_3d, TAU_OG_3d, W0_OG_3d, COSB_OG_3d,
											atm.surf_reflect, ubar0,ubar1,cos_theta, F0PI,
											single_phase,multi_phase,
											frac_a,frac_b,frac_c,constant_back,constant_forward)

	#now compress everything based on the weights 
	if  ('reflected' in calculation) & ('thermal' not in calculation):
		albedo = compress_disco(nwno, cos_theta, xint_at_top, gweight, tweight,F0PI)
		returns = (wno, albedo)

	elif ('reflected' not in calculation) & ('thermal' in calculation):
		thermal = compress_thermal(nwno,ubar1_therm, flux_at_top, gweight_therm, tweight_therm)
		fpfs_thermal = thermal/(opacityclass.unshifted_stellar_spec)*(atm.planet.radius/radius_star)**2.0
		returns = wno,fpfs_thermal#,thermal

	elif ('reflected' in calculation) & ('thermal' in calculation):
		albedo = compress_disco(nwno, cos_theta, xint_at_top, gweight, tweight,F0PI)
		thermal = compress_thermal(nwno,ubar1_therm, flux_at_top, gweight_therm, tweight_therm)
		fpfs_thermal = thermal/(opacityclass.unshifted_stellar_spec)*(atm.planet.radius/radius_star)**2.0
		returns = wno,albedo, fpfs_thermal#,thermal

	if full_output:	
		#add full solution and latitude and longitudes to the full output
		#atm.flux_at_top = flux_at_top
		atm.xint_at_top = xint_at_top
		atm.latitude = lat
		atm.longitude = lon
		return wno, albedo , atm.as_dict()
	else: 
		return returns

def opannection(filename_db = None, raman_db = None):
	"""
	Sets up database connection to opacities. 
	"""

	inputs = json.load(open(os.path.join(__refdata__,'config.json')))

	if isinstance(filename_db,type(None) ): filename_db = os.path.join(__refdata__, 'opacities', inputs['opacities']['files']['opacity'])
	if isinstance(raman_db,type(None) ): raman_db = os.path.join(__refdata__, 'opacities', inputs['opacities']['files']['raman'])

	opacityclass=RetrieveOpacities(
				filename_db, 
				raman_db
				)
	return opacityclass

class inputs():
	"""Class to setup planet to run

	Parameters
	----------
	continuum_db: str 
		(Optional)Filename that points to the HDF5 contimuum database (see notebook on swapping out opacities).
		This database pointer will also define the wavelength grid to run on. So, wavenumber grid 
		should be specified in the database as an attribute!
		Default will pull the zenodo defualt opacities 
	molecular_db: str 
		(Optional)Filename that points to the HDF5 molecular database (see notebook on swapping out opacities).
		This database pointer will also define the wavelength grid to run on. So, wavenumber grid 
		should be specified in the database as an attribute!
		Default will pull the zenodo defualt opacities 
	raman_db: str 
		(Optional)Filename that points to the raman scattering database. Default is the text file from Ocklopcic+2018 paper 
		on exoplanet raman scattering. 

	Attributes
	----------
	phase_angle() : set phase angle
	gravity() : set gravity
	star() : set stellar spec
	atmosphere() : set atmosphere composition and PT
	clouds() : set cloud profile
	approx()  : set approximation
	spectrum() : create spectrum
	"""
	def __init__(self, chemeq=False):#continuum_db = None, molecular_db = None, raman_db = None):

		self.inputs = json.load(open(os.path.join(__refdata__,'config.json')))

		#if runnng chemical equilibrium, need to load chemeq grid
		if chemeq: self.chemeq_pic = pk.load(open(os.path.join(__refdata__,'chem_full.pic'),'rb'), encoding='latin1')


	def phase_angle(self, phase=0,num_gangle=10, num_tangle=10):
		"""Define phase angle and number of gauss and tchebychev angles to compute. 
		
		phase : float,int
			Phase angle in radians 
		num_gangle : int 
			Number of Gauss angles to integrate over facets (Default is 10).
			 Higher numbers will slow down code. 
		num_tangle : int 
			Number of Tchebyshev angles to integrate over facets (default is 10)
		"""
		if (num_gangle < 2 ) or (num_tangle < 2 ): raise Exception("length of gangle and tangle must be > than 2")
		self.inputs['phase_angle'] = phase
		self.inputs['disco']['num_gangle'] = int(num_gangle)
		self.inputs['disco']['num_tangle'] = int(num_tangle)

	def gravity(self, gravity=None, gravity_unit=None, 
		              radius=None, radius_unit=None, 
		              mass = None, mass_unit=None):
		"""
		Get gravity based on mass and radius, or gravity inputs 

		Parameters
		----------
		gravity : float 
			(Optional) Gravity of planet 
		gravity_unit : astropy.unit
			(Optional) Unit of Gravity
		radius : float 
			(Optional) radius of planet MUST be specified for thermal emission!
		radius_unit : astropy.unit
			(Optional) Unit of radius
		mass : float 
			(Optional) mass of planet 
		mass_unit : astropy.unit
			(Optional) Unit of mass	
		"""
		if (mass is not None) and (radius is not None):
			m = (mass*mass_unit).to(u.g)
			r = (radius*radius_unit).to(u.cm)
			g = (c.G * m /  (r**2)).value
			self.inputs['planet']['radius'] = r.value
			self.inputs['planet']['radius_unit'] = 'cm'
			self.inputs['planet']['mass'] = m.value
			self.inputs['planet']['mass_unit'] = 'g'
			self.inputs['planet']['gravity'] = g
			self.inputs['planet']['gravity_unit'] = 'cm/(s**2)'
		elif gravity is not None:
			g = (gravity*gravity_unit).to('cm/(s**2)')
			g = g.value
			self.inputs['planet']['gravity'] = g
			self.inputs['planet']['gravity_unit'] = 'cm/(s**2)'
			self.inputs['planet']['radius'] = np.nan
			self.inputs['planet']['radius_unit'] = 'Radius not specified'
		else: 
			raise Exception('Need to specify gravity or radius and mass + additional units')

	def star(self, opannection,temp=None, metal=None, logg=None ,radius = None, radius_unit=None,
		database='ck04models',filename=None, w_units=None, f_units=None):
		"""
		Get the stellar spectrum using pysynphot and interpolate onto a much finer grid than the 
		planet grid. 

		Parameters
		----------
		opannection : class picaso.RetrieveOpacities
			This is the opacity class and it's needed to get the correct wave info and raman scattering cross sections
		temp : float 
			Teff of the stellar model 
		metal : float 
			Metallicity of the stellar model 
		logg : float 
			Logg cgs of the stellar model
		radius : float 
			Radius of the star 
		radius_unit : astropy.unit
			Any astropy unit (e.g. `radius_unit=astropy.unit.Unit("R_sun")`)
		database : str 
			(Optional)The database to pull stellar spectrum from. See documentation for pysynphot. 
		filename : str 
			(Optional) Upload your own stellar spectrum. File format = two column white space (wave, flux)
		wunits : str 
			(Optional) Used for stellar file wave units 
		funits : str 
			(Optional) Used for stellar file flux units 
		"""
		#most people will just upload their thing from a database
		if (not isinstance(radius, type(None))):
			r = (radius*radius_unit).to(u.cm).value
			radius_unit='cm'
		else :
			r = np.nan
			radius_unit = "Radius not supplied"

		if (not isinstance(temp, type(None))):
			sp = psyn.Icat(database, temp, metal, logg)
			sp.convert("um")
			sp.convert('flam') 
			wno_star = 1e4/sp.wave[::-1] #convert to wave number and flip
			flux_star = sp.flux[::-1]*1e8	 #flip here and convert to ergs/cm3/s to get correct order

		#but you can also upload a stellar spec of your own 
		elif (not isinstance(filename),type(none)):
			star = np.genfromtxt(filename, dtype=(float, float), names='w, f')
			flux = star['f']
			wave = star['w']
			#sort if not in ascending order 
			sort = np.array([wave,flux]).T
			sort= sort[sort[:,0].argsort()]
			wave = sort[:,0]
			flux = sort[:,1] 
			if w_unit == 'um':
				WAVEUNITS = 'um' 
			elif w_unit == 'nm':
				WAVEUNITS = 'nm'
			elif w_unit == 'cm' :
				WAVEUNITS = 'cm'
			elif w_unit == 'Angs' :
				WAVEUNITS = 'angstrom'
			elif w_unit == 'Hz' :
				WAVEUNITS = 'Hz'
			else: 
				raise Exception('Stellar units are not correct. Pick um, nm, cm, hz, or Angs')        

			#http://www.gemini.edu/sciops/instruments/integration-time-calculators/itc-help/source-definition
			if f_unit == 'Jy':
				FLUXUNITS = 'jy' 
			elif f_unit == 'FLAM' :
				FLUXUNITS = 'FLAM'
			elif f_unit == 'erg/cm2/s/Hz':
				flux = flux*1e23
				FLUXUNITS = 'jy' 
			else: 
				raise Exception('Stellar units are not correct. Pick FLAM or Jy or erg/cm2/s/Hz')

			sp = psyn.ArraySpectrum(wave, flux, waveunits=WAVEUNITS, fluxunits=FLUXUNITS)        #Convert evrything to nanometer for converstion based on gemini.edu  
			sp.convert("um")
			sp.convert('flam') #ergs/cm2/s/ang
			wno_star = 1e4/sp.wave[::-1] #convert to wave number and flip
			flux_star = sp.flux[::-1]*1e8 #flip and convert to ergs/cm3/s here to get correct order			


		wno_planet = opannection.wno
		max_shift = np.max(wno_planet)+6000 #this 6000 is just the max raman shift we could have 
		min_shift = np.min(wno_planet) -2000 #it is just to make sure we cut off the right wave ranges

		#this adds stellar shifts 'self.raman_stellar_shifts' to the opacity class
		#the cross sections are computed later 
		if self.inputs['approx']['raman'] == 0: 
			#do a fail safe to make sure that star is on a fine enough grid for planet case 
			fine_wno_star = np.linspace(min_shift, max_shift, len(wno_planet)*5)
			fine_flux_star = np.interp(fine_wno_star,wno_star, flux_star)
			opannection.compute_stellar_shits(fine_wno_star, fine_flux_star)
		else :
			fine_wno_star = wno_planet
			fine_flux_star = np.interp(wno_planet,wno_star, flux_star)	
			opannection.unshifted_stellar_spec =fine_flux_star

		self.inputs['star']['database'] = database
		self.inputs['star']['temp'] = temp
		self.inputs['star']['logg'] = logg
		self.inputs['star']['metal'] = metal
		self.inputs['star']['radius'] = r 
		self.inputs['star']['radius_unit'] = radius_unit 
		self.inputs['star']['flux'] = fine_flux_star 
		self.inputs['star']['wno'] = fine_wno_star 

	def atmosphere(self, df=None, filename=None, exclude_mol=None, **pd_kwargs):
		"""
		Builds a dataframe and makes sure that minimum necessary parameters have been suplied.
		Sets number of layers in model.  

		Parameters
		----------
		df : pandas.DataFrame or dict
			(Optional) Dataframe with volume mixing ratios and pressure, temperature profile. 
			Must contain pressure (bars) at least one molecule
		filename : str 
			(Optional) Filename with pressure, temperature and volume mixing ratios.
			Must contain pressure at least one molecule
		exclude_mol : list of str 
			(Optional) List of molecules to ignore from file
		pd_kwargs : kwargs 
			Key word arguments for pd.read_csv to read in supplied atmosphere file 
		"""

		if not isinstance(df, type(None)):
			if ((not isinstance(df, dict )) & (not isinstance(df, pd.core.frame.DataFrame ))): 
				raise Exception("df must be pandas DataFrame or dictionary")
			else:
				self.nlevel=df.shape[0] 
		elif not isinstance(filename, type(None)):
			df = pd.read_csv(filename, **pd_kwargs)
			self.nlevel=df.shape[0] 
		elif not isinstance(pt_params, type(None)): 
			self.inputs['atmosphere']['pt_params'] = pt_params
			self.nlevel=61 #default n of levels for parameterization
			raise Exception('Pt parameterization not in yet')

		if 'pressure' not in df.keys(): 
			raise Exception("Check column names. `pressure` must be included.")

		if (('temperature' not in df.keys()) and (isinstance(pt_params, type(None)))):
			raise Exception("`temperature` not specified as a column/key name, and `pt_params` for a parameterized PT profile was not supplied. Please make sure to use one or the other.")

		if not isinstance(exclude_mol, type(None)):
			df = df.drop(exclude_mol, axis=1)
			self.inputs['atmosphere']['exclude_mol'] = exclude_mol

		self.inputs['atmosphere']['profile'] = df.sort_values('pressure').reset_index(drop=True)

	def chemeq(self, CtoO, Met, P=None,T=None):
		"""
		This interpolates from a precomputed grid of CEA runs (run by M.R. Line)

		Parameters
		----------
		P : array
			Pressure (bars)
		T : array
			Temperature (K)
		CtoO : float
			log C to O ratio (log solar = -0.26)
		Met : float 
			log Metallicity relative to solar (solar = 0 (log10(1) ))
		"""
		 
		P, T = self.inputs['atmosphere']['profile']['pressure'].values,self.inputs['atmosphere']['profile']['temperature'].values
		
		T[T<400] = 400
		T[T>2800] = 2800

		logCtoO, logMet, Tarr, logParr, gases=self.chemeq_pic
		assert Met < 10**np.max(logMet), 'Metallicity entered is higher than the max of the grid: M/H = '+ str(np.max(10**logMet))
		assert CtoO < 10**np.max(logCtoO), 'C/O ratio entered is higher than the max of the grid: C/O = '+ str(np.max(10**logCtoO))
		assert Met > 10**np.min(logMet), 'Metallicity entered is higher than the max of the grid: M/H = '+ str(np.min(10**logMet))
		assert CtoO > 10**np.min(logCtoO), 'C/O ratio entered is higher than the max of the grid: C/O = '+ str(np.min(10**logCtoO))

		loggas=np.log10(gases)
		Ngas = loggas.shape[3]
		gas=np.zeros((Ngas,len(P)))
		for j in range(Ngas):
			gas_to_interp=loggas[:,:,:,j,:]
			IF=RegularGridInterpolator((logCtoO, logMet, np.log10(Tarr),logParr),gas_to_interp,bounds_error=False)
			for i in range(len(P)):
				gas[j,i]=10**IF(np.array([np.log10(CtoO), np.log10(Met), np.log10(T[i]), np.log10(P[i])]))
		H2Oarr, CH4arr, COarr, CO2arr, NH3arr, N2arr, HCNarr, H2Sarr,PH3arr, C2H2arr, C2H6arr, Naarr, Karr, TiOarr, VOarr, FeHarr, Harr,H2arr, Hearr, mmw=gas

		df = pd.DataFrame({'H2O': H2Oarr, 'CH4': CH4arr, 'CO': COarr, 'CO2': CO2arr, 'NH3': NH3arr, 
			               'N2' : N2arr, 'HCN': HCNarr, 'H2S': H2Sarr, 'PH3': PH3arr, 'C2H2': C2H2arr, 
			               'C2H6' :C2H6arr, 'Na' : Naarr, 'K' : Karr, 'TiO': TiOarr, 'VO' : VOarr, 
			               'Fe': FeHarr,  'H': Harr, 'H2' : H2arr, 'He' : Hearr, 'temperature':T, 
			               'pressure': P})
		self.inputs['atmosphere']['profile'] = df
		return 


	def guillot_pt(self, Teq, T_int, logg1, logKir, alpha=0.5,nlevel=61):
		"""
		Creates temperature pressure profile given parameterization in Guillot 2010 TP profile
		called in fx()

		Parameters
		----------
		Teq : float 
		    equilibrium temperature 
		T_int : float 
		    Internal temperature, if low (100) currently set to 100 for everything  
		kv1 : float 
		    see parameterization Guillot 2010 (10.**(logg1+logKir))
		kv2 : float
		    see parameterization Guillot 2010 (10.**(logg1+logKir))
		kth : float
		    see parameterization Guillot 2010 (10.**logKir)
		alpha : float 
		    set to 0.5
		nlevel : int
			Number of atmospheric layers
		    
		Returns
		-------
		T : numpy.array 
		    Temperature grid 
		P : numpy.array
		    Pressure grid
		        
		"""
		kv1, kv2 =10.**(logg1+logKir),10.**(logg1+logKir)
		kth=10.**logKir

		Teff = T_int
		f = 1.0  # solar re-radiation factor
		A = 0.0  # planetary albedo
		g0 = self.inputs['planet']['gravity']/100.0 #cm/s2 to m/s2

		# Compute equilibrium temperature and set up gamma's
		T0 = Teq
		gamma1 = kv1/kth #Eqn. 25
		gamma2 = kv2/kth

		# Initialize arrays
		logtau =np.arange(-10,20,.1)
		tau =10**logtau

		#computing temperature
		T4ir = 0.75*(Teff**(4.))*(tau+(2.0/3.0))
		f1 = 2.0/3.0 + 2.0/(3.0*gamma1)*(1.+(gamma1*tau/2.0-1.0)*sp.exp(-gamma1*tau))+2.0*gamma1/3.0*(1.0-tau**2.0/2.0)*special.expn(2.0,gamma1*tau)
		f2 = 2.0/3.0 + 2.0/(3.0*gamma2)*(1.+(gamma2*tau/2.0-1.0)*sp.exp(-gamma2*tau))+2.0*gamma2/3.0*(1.0-tau**2.0/2.0)*special.expn(2.0,gamma2*tau)
		T4v1=f*0.75*T0**4.0*(1.0-alpha)*f1
		T4v2=f*0.75*T0**4.0*alpha*f2
		T=(T4ir+T4v1+T4v2)**(0.25)
		P=tau*g0/(kth*0.1)/1.E5
		self.nlevel=nlevel 
		logP = np.linspace(-6.8,1.5,nlevel)
		newP = 10.0**logP
		T = np.interp(logP,np.log10(P),T)

		self.inputs['atmosphere']['profile']  = pd.DataFrame({'temperature': T, 'pressure':newP})

		# Return TP profile
		return self.inputs['atmosphere']['profile'] 

	def atmosphere_3d(self, dictionary=None, filename=None, exclude_mol=None, **pd_kwargs):
		"""
		Builds a dataframe and makes sure that minimum necessary parameters have been suplied. 

		Parameters
		----------
		df : pandas.DataFrame
			(Optional) Dataframe with volume mixing ratios and pressure, temperature profile. 
			Must contain pressure (bars) at least one molecule
		filename : str 
			(Optional) Pickle filename that is a dictionary structured: 
			dictionary[float(lat)][float(lon)] = pd.DataFrame({'pressure':[], 'temperature': [],
			'H2O':[]....}). As well as df['phase_angle'] = 0 (in radians).
			Therefore, the latitude and longitude keys should be floats, any other meta data 
			should be  represented by a string.
			Must contain pressure at least one molecule
		exclude_mol : list of str 
			(Optional) List of molecules to ignore from file
		pd_kwargs : kwargs 
			Key word arguments for pd.read_csv to read in supplied atmosphere file 
		"""
		if not isinstance(dictionary, type(None)):

			if (not isinstance(dictionary, dict )): 
				raise Exception("df must be pandas DataFrame or dictionary")
			else:
				self.nlevel=df.shape[0] 

		elif not isinstance(filename, type(None)):
			df3d = pk.load(open(filename,'rb'))

		df3d_phase = 	df3d['phase_angle']

		lats = list(i for i in df3d.keys() if isinstance(i,float))
		df3d_nt = len(lats)
		df3d_ng =  len(df3d[lats[0]].keys())
		np.testing.assert_almost_equal(self.inputs['phase_angle'] ,df3d_phase, decimal=1,err_msg='Phase Angles not the same as 3d input file', verbose=True)
		assert self.inputs['disco']['num_gangle'] == int(df3d_nt), 'Number of gauss angles does not match creation of 3d input file'
		assert self.inputs['disco']['num_tangle']  == int(df3d_ng), 'Number of Tchebyshev angles does not match creation of 3d input file'

		self.inputs['atmosphere']['profile'] = df3d
		
		return

	def clouds(self, filename = None, g0=None, w0=None, opd=None,p=None, dp=None,df =None,**pd_kwargs):
		"""
		Cloud specification for the model. Clouds are parameterized by a single scattering albedo (w0), 
		an assymetry parameter (g0), and a total extinction per layer (opd).

		g0,w0, and opd are both wavelength and pressure dependent. Our cloud models come 
		from eddysed. Their output looks like this (where level 1=TOA, and wavenumber1=Smallest number)

		level wavenumber opd w0 g0
		1.	 1.   ... . .
		1.	 2.   ... . .
		1.	 3.   ... . .
		.	  .	... . .
		.	  .	... . .
		1.	 M.   ... . .
		2.	 1.   ... . .
		.	  .	... . .
		N.	 .	... . .

		If you are creating your own file you have to make sure that you have a 
		**pressure** (bars) and **wavenumber**(inverse cm) column. We will use this to make sure that your cloud 
		and atmospheric profiles are on the same grid. **If there is no pressure or wavelength parameter
		we will assume that you are on the same grid as your atmospheric input, and on the 
		eddysed wavelength grid! **

		Users can also input their own fixed cloud parameters, by specifying a single value 
		for g0,w0,opd and defining the thickness and location of the cloud. 

		Parameters
		----------
		filename : str 
			(Optional) Filename with info on the wavelength and pressure-dependent single scattering
			albedo, asymmetry factor, and total extinction per layer. Input associated pd_kwargs 
			so that the resultant output has columns named : `g0`, `w0` and `opd`. If you are not 
			using the eddysed output, you will also need a `wavenumber` and `pressure` column in units 
			of inverse cm, and bars. 
		g0 : float, list of float
			(Optional) Asymmetry factor. Can be a single float for a single cloud. Or a list of floats 
			for two different cloud layers 
		w0 : list of float 
			(Optional) Single Scattering Albedo. Can be a single float for a single cloud. Or a list of floats 
			for two different cloud layers 		
		opd : list of float 
			(Optional) Total Extinction in `dp`. Can be a single float for a single cloud. Or a list of floats 
			for two different cloud layers 
		p : list of float 
			(Optional) Center location of cloud deck (bars). Can be a single float for a single cloud. Or a list of floats 
			for two different cloud layers 
		dp : list of float 
			(Optional) Total thickness cloud deck (bars). Can be a single float for a single cloud or a list of floats 
			for two different cloud layers 
			Cloud will span 10**(np.log10(p) +- np.log10(dp)/2)
		df : pd.DataFrame, dict
			(Optional) Same as what would be included in the file, but in DataFrame or dict form
		"""

		#first complete options if user inputs dataframe or dict 
		if (not isinstance(filename, type(None)) & isinstance(df, type(None))) or (isinstance(filename, type(None)) & (not isinstance(df, type(None)))):

			if not isinstance(filename, type(None)):
				df = pd.read_csv(filename, **pd_kwargs)

			cols = df.keys()

			assert 'g0' in cols, "Please make sure g0 is a named column in cld file"
			assert 'w0' in cols, "Please make sure w0 is a named column in cld file"
			assert 'opd' in cols, "Please make sure opd is a named column in cld file"

			#CHECK SIZES

			#if it's a user specified pressure and wavenumber
			if (('pressure' in cols) & ('wavenumber' in cols)):
				df = df.sort_values(['wavenumber','pressure']).reset_index(drop=True)
				self.inputs['clouds']['wavenumber'] = df['wavenumber'].unique()
				nwave = len(self.inputs['clouds']['wavenumber'])
				nlayer = len(df['pressure'].unique())
				assert df.shape[0] == (self.nlevel-1)*nwave, "There are {0} rows in the df, which does not equal {1} layers previously specified x {2} wave pts".format(df.shape[0], self.nlevel-1, nwave) 
			
			#if its eddysed, make sure there are 196 wave points 
			else: 
				assert df.shape[0] == (self.nlevel-1)*196, "There are {0} rows in the df, which does not equal {1} layers x 196 eddysed wave pts".format(df.shape[0], self.nlevel-1) 
				
				self.inputs['clouds']['wavenumber'] = get_cld_input_grid('wave_EGP.dat')

			#add it to input
			self.inputs['clouds']['profile'] = df

		#first make sure that all of these have been specified
		elif None in [g0, w0, opd, p,dp]:
			raise Exception("Must either give dataframe/dict, OR a complete set of g0, w0, opd,p,dp to compute cloud profile")
		else:
			pressure_level = self.inputs['atmosphere']['profile']['pressure'].values
			pressure = np.sqrt(pressure_level[1:] * pressure_level[0:-1])#layer

			w = get_cld_input_grid('wave_EGP.dat')

			self.inputs['clouds']['wavenumber'] = w

			pressure_all =[]
			for i in pressure: pressure_all += [i]*len(w)
			wave_all = list(w)*len(pressure)

			df = pd.DataFrame({'pressure':pressure_all,
								'wavenumber': wave_all })


			zeros=np.zeros(196*(self.nlevel-1))

			#add in cloud layers 
			df['g0'] = zeros
			df['w0'] = zeros
			df['opd'] = zeros
			#loop through all cloud layers and set cloud profile
			for ig, iw, io , ip, idp in zip(g0,w0,opd,p,dp):
				minp = 10**(np.log10(ip) - np.log10(idp/2))
				maxp = 10**(np.log10(ip) + np.log10(idp/2))
				df.loc[((df['pressure'] >= minp) & (df['pressure'] <= maxp)),'g0']= ig
				df.loc[((df['pressure'] >= minp) & (df['pressure'] <= maxp)),'w0']= iw
				df.loc[((df['pressure'] >= minp) & (df['pressure'] <= maxp)),'opd']= io

			self.inputs['clouds']['profile'] = df

	def approx(self,single_phase='TTHG_ray',multi_phase='N=2',delta_eddington=True,raman='oklopcic',
				tthg_frac=[1,-1,2], tthg_back=-0.5, tthg_forward=1):
		"""
		This function sets all the default approximations in the code. It transforms the string specificatons
		into a number so that they can be used in numba nopython routines. 

		For `str` cases such as `TTHG_ray` users see all the options by using the function `single_phase_options`
		or `multi_phase_options`, etc. 

		single_phase : str 
			Single scattering phase function approximation 
		multi_phase : str 
			Multiple scattering phase function approximation 
		delta_eddington : bool 
			Turns delta-eddington on and off
		raman : str 
			Uses various versions of raman scattering 
		tthg_frac : list 
			Functional of forward to back scattering with the form of polynomial :
			tthg_frac[0] + tthg_frac[1]*g_b^tthg_frac[2]
			See eqn. 6 in picaso paper 
		tthg_back : float 
			Back scattering asymmetry factor gf = g_bar*tthg_back
		tthg_forward : float 
			Forward scattering asymmetry factor gb = g_bar * tthg_forward 
		"""

		self.inputs['approx']['single_phase'] = single_phase_options(printout=False).index(single_phase)
		self.inputs['approx']['multi_phase'] = multi_phase_options(printout=False).index(multi_phase)
		self.inputs['approx']['delta_eddington'] = delta_eddington
		self.inputs['approx']['raman'] =  raman_options().index(raman)

		if isinstance(tthg_frac, (list, np.ndarray)):
			if len(tthg_frac) == 3:
				self.inputs['approx']['TTHG_params']['fraction'] = tthg_frac
			else:
				raise Exception('tthg_frac should be of length=3 so that : tthg_frac[0] + tthg_frac[1]*g_b^tthg_frac[2]')
		else: 
			raise Exception('tthg_frac should be a list or ndarray of length=3')

		self.inputs['approx']['TTHG_params']['constant_back'] = tthg_back
		self.inputs['approx']['TTHG_params']['constant_forward']=tthg_forward


	def spectrum(self,opacityclass,dimension = '1d', calculation='reflected', full_output=False, plot_opacity= False):
		"""Run Spectrum"""
		if ('thermal' in calculation) and (np.isnan(self.inputs['star']['radius']) or np.isnan(self.inputs['planet']['radius'])):
			raise Exception("Stellar or Planet radius not supplied but thermal flux was requested. See options in `star()` `gravity()`")
			
		return picaso(self, opacityclass,dimension=dimension,calculation=calculation,
			full_output=full_output, plot_opacity=plot_opacity)


def jupiter_pt():
	"""Function to get Jupiter's PT profile"""
	return os.path.join(__refdata__, 'base_cases','jupiter.pt')
def jupiter_cld():
	"""Function to get rough Jupiter Cloud model with fsed=3"""
	return os.path.join(__refdata__, 'base_cases','jupiterf3.cld')
def single_phase_options(printout=True):
	"""Retrieve all the options for direct radation"""
	if printout: print("Can also set functional form of forward/back scattering in approx['TTHG_params']")
	return ['cahoy','OTHG','TTHG','TTHG_ray']
def multi_phase_options(printout=True):
	"""Retrieve all the options for multiple scattering radiation"""
	if printout: print("Can also set delta_eddington=True/False in approx['delta_eddington']")
	return ['N=2','N=1']
def raman_options():
	"""Retrieve options for raman scattering approximtions"""
	return ["oklopcic","pollack","none"]




import json
from .disco import get_angles, compute_disco
import pandas as pd
import numpy as np
from scipy.spatial import cKDTree
import astropy.units as u
def rebin_mitgcm(ng, nt, phase_angle, input_file, output_file,p_unit='Pa', run_chem=False, MH=None, CtoO=None):
	"""
	Rebin GCM grid to a smaller grid. Function does not yet work!! 

	Parameters 
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle (radians)
	input_file : str
		PT input file you want to create the 3d grid from. Currently takes in a single 1d PT profile.
		Must have labeled, whitespace delimeted columns. In the columns, must have `temperature`, 
		`pressure`, and at least some molecular species (which are ALL case-sensitive)
	output_file : str
		output file location
	p_unit : str
		Pressure Unit (default Pascal)
	run_chem : bool 
		(Optional) runs chemistry and adds to json file

	TO DO
	-----
	- add run chemistry to get 3d chem
	"""	

	gangle,gweight,tangle,tweight = get_angles(ng,nt)
	ubar0, ubar1, cos_theta ,lat_picaso,lon_picaso = compute_disco(ng,nt, gangle, tangle, phase_angle)    


	infile = open(input_file,'r')

	temp = infile.readline().split()
	nlon = int(temp[0])
	nlat = int(temp[1])
	nz = int(temp[2])
	total_pts = nlon * nlat
	all_lon = np.zeros(total_pts) # 128 x 64
	all_lat = np.zeros(total_pts)
	p = np.zeros(nz)
	t = np.zeros((nlon,nlat,nz))
	kzz = np.zeros((nlon,nlat,nz))

	# skip line of header
	temp = infile.readline()
	ctr = -1

	for ilon in range(0,nlon):
		for ilat in range(0,nlat):
			ctr += 1

			#skip blank line
			temp = infile.readline()
			temp = infile.readline().split()

			all_lon[ctr] = float(temp[0])
			all_lat[ctr] = float(temp[1])

			# read in data for each grid point
			for iz in range(0,nz):
				temp = infile.readline().split()
				p[iz] = float(temp[0])
				t[ilon,ilat,iz] = float(temp[1])
				kzz[ilon,ilat,iz] = float(temp[2])

	lon = np.unique(all_lon)
	lat = np.unique(all_lat)

	lon2d, lat2d = np.meshgrid(lon_picaso, lat_picaso)
	lon2d = lon2d.flatten()
	lat2d = lat2d.flatten()

	xs, ys, zs = lon_lat_to_cartesian(all_lon,all_lat)
	xt, yt, zt = lon_lat_to_cartesian(lon2d,lat2d)

	tree = cKDTree(list(zip(xs,ys,zs)))
	nn = int(total_pts / (ng*nt))
	d,inds = tree.query(list(zip(xt,yt,zt)),k=nn)

	new_t = np.zeros((ng*nt,nz))
	new_kzz = np.zeros((ng*nt,nz))
	for iz in range(0,nz):
	    new_t[:,iz] = np.sum(t[:,:,iz].flatten()[inds],axis=1)/nn
	    new_kzz[:,iz] = np.sum(kzz[:,:,iz].flatten()[inds],axis=1)/nn


	df = {} 
	df={i:{} for i in lat_picaso}


	outfile = open(output_file,'w')
	for ipos in range(0,ng*nt):

		#make sure units are always bars
		p = (p*u.Unit(p_unit)).to(u.bar).value

		if run_chem:
			case = jdi.inputs()
			case.inputs['atmosphere']['profile'] = pd.DataFrame({'temperature':new_t[ipos,0:nz], 'pressure':p, 'kzz':new_kzz[ipos,0:nz]})
			case.chemeq(CtoO, MH)
			data = case.inputs['atmosphere']['profile']
		else : 
			data = pd.DataFrame({'temperature':new_t[ipos,0:nz], 'pressure':p, 'kzz':new_kzz[ipos,0:nz]})

		df[lat2d[ilat]][lon2d[ipos]] = data

	df['phase_angle'] = phase_angle

def make_3d_pt_input(ng,nt,phase_angle,input_file,output_file,**kwargs):
	"""
	Program to create 3d PT input. Used to feed GCM input into disco ball.

	Parameters
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle
	input_file : str
		PT input file you want to create the 3d grid from. Currently takes in a single 1d PT profile.
		Must have labeled, whitespace delimeted columns. In the columns, must have `temperature`, 
		`pressure`, and at least some molecular species (which are ALL case-sensitive)
	output_file : str
		Output file location

	Returns
	-------
	Creates output file. No other returns.

	Examples
	--------
	Basic use of creating a 3D file: 

	>>> phase_angle = 0
	>>> number_gauss_angles = 2
	>>> number_tcheby_angles = 3
	>>> phase_angle = 0
	>>> input_file=jdi.jupiter_pt()
	>>> output_file='jupiter3D.hdf5'
	>>>make_3d_pt_input(number_gauss_angles, number_tcheby_angles,
	>>>         phase_angle, input_file,output_file)
	
	Can also use `**kwargs` to control how pandas reads in input file 

	>>>make_3d_pt_input(number_gauss_angles, number_tcheby_angles,
	>>>         phase_angle, input_file,output_file,usecols=[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,20,21])

	"""

	h5db = h5py.File(output_file,'w')
	#get geometry
	gangle,gweight,tangle,tweight = disco.get_angles(ng, nt)
	ubar0, ubar1, cos_theta,lat,lon = disco.compute_disco(ng, nt, gangle, tangle, phase_angle)



	first = True
	dset = {}
	for g in gangle:
		for t in tangle:
			data = pd.read_csv(input_file, delim_whitespace=True,**kwargs)
			if first:
				dset['header'] = ','.join(list(data.keys()))
				first = False
			dset[str(g)][str(t)] = data

def make_3d_cld_input(ng,nt,phase_angle,input_file,output_file, lat_range=None, lon_range=None,rand_coverage=1):
	"""
	Program to create 3d CLOUD input. Used to feed GCM input into disco ball.

	Parameters
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle
	input_file : str
		CLD input file you want to create the 3d grid from. Currently takes in a single 1d CLD profile.
		Input file must be structured with the following columns in the correct order: ['lvl', 'wv','opd','g0','w0','sigma']
		input file 
	output_file : str
		Output file location
	lat_range : list
		(Optional)Range of latitudes to exclude in the CLD file
	lon_range : list
		(Optional)Range of longitudes to exclud in the CLD file
	rand_coverage : float 
		(Optional)Fractional cloud coverage. rand_cov=1 introduces full cloud coverage. rand_cov=0 has no
		coverage. 
	**kwargs : dict 
		Key word arguments used for `panads.read_csv`

	Returns
	-------
	Creates output file. No other returns.

	Examples
	--------
	Basic use of creating a 3D file: 
		
	>>> phase_angle = 0
	>>> number_gauss_angles = 2
	>>> number_tcheby_angles = 3
	>>> phase_angle = 0
	>>> input_file=jdi.jupiter_cld()
	>>> output_file='jupiter3D.hdf5'
	>>>make_3d_pt_input(number_gauss_angles, number_tcheby_angles,
	>>>         phase_angle, input_file,output_file)
	"""

	h5db = h5py.File(output_file,'w')
	#get geometry
	gangle,gweight,tangle,tweight = disco.get_angles(ng, nt)
	ubar0, ubar1, cos_theta,lat,lon = disco.compute_disco(ng, nt, gangle, tangle, phase_angle)
	first = True

	for g, lg in zip(gangle,lon):
		for t, lt in zip(tangle, lat):
			data = pd.read_csv(input_file, delim_whitespace = True,
					header=None, skiprows=1, names = ['lvl', 'wv','opd','g0','w0','sigma'],
					dtype='f8')
			if not isinstance(lat_range,type(None)):
				if (lt > np.min(lat_range)) and (lt<np.max(lat_range)):
					data = data*0
			if not isinstance(lon_range,type(None)):
				if (lg > np.min(lon_range)) and (lg < np.max(lon_range)):
					data = data*0
			if np.random.rand() > rand_coverage:
				data = data*0

			if first:
				h5db.attrs['header'] = ','.join(list(data.keys()))
				first = False
			dset = h5db.create_dataset(str(g)+'/'+str(t), data=data, chunks=True)

def lon_lat_to_cartesian(lon, lat, R = 1):
	"""
	Calculates lon, lat coordinates of a point on a sphere with
	radius R

	Parameters 
	----------
	lon : array
		Longitude
	lat : array 
		Latitude 
	R : float	
		radius of sphere 

	Returns
	-------
	x, y, z 
	"""
	x =  R * np.cos(lat_r) * np.cos(lon_r)
	y = R * np.cos(lat_r) * np.sin(lon_r)
	z = R * np.sin(lat_r)
	return x,y,z



from numba import jit, vectorize
from numpy import exp, zeros, where, sqrt, cumsum , pi, outer

@jit(nopython=True, cache=True)
def get_flux_toon(nlevel, wno, nwno, tau, dtau, w0, cosbar, surf_reflect, ubar0, F0PI):
	"""
	Warning
	-------
	Discontinued function. See `get_flux_geom_1d` and `get_flux_geom_3d`.

	Parameters
	----------
	nlevel : int 
		Number of levels in the model 
	wno : array of float 
		Wave number grid in cm -1 
	nwno : int 
		Number of wave points
	dtau : ndarray of float
		This is the opacity contained within each individual layer (defined at midpoints of "levels")
		Dimensions=# layer by # wave
	w0 : ndarray of float 
		This is the single scattering albedo, from scattering, clouds, raman, etc 
		Dimensions=# layer by # wave
	cosbar : ndarray of float 
		This is the asymmetry factor 
		Dimensions=# layer by # wave
	surf_reflect : float 
		Surface reflectivity 
	ubar0 : float 
		Cosine of the incident angle 
	F0PI : array 
		Downward incident solar radiation


	Returns
	-------
	flux_up and flux_down through each layer as a function of wavelength 

	Todo
	----
	- Replace detla-function adjustment with better approximation (e.g. Cuzzi)
	- F0PI Solar flux shouldn't always be 1.. Follow up to make sure that this isn't a bad 
		  hardwiring to solar 
	
	Examples
	--------
	
	>>> flux_plus, flux_minus  = fluxes.get_flux_toon(atm.c.nlevel, wno,nwno,
													tau_dedd,dtau_dedd, w0_dedd, cosb_dedd, surf_reflect, ubar0, F0PI)


	"""
	nlayer = nlevel - 1 

	#initially used for the downard flux calc
	#ubar0 = 0.5 
	#F0PI = np.zeros(wno) + 1.0 

	#First thing to do is to use the delta function to icorporate the forward 
	#peak contribution of scattering by adjusting optical properties such that 
	#the fraction of scattered energy in the forward direction is removed from 
	#the scattering parameters 

	#Joseph, J.H., W. J. Wiscombe, and J. A. Weinman, 
	#The Delta-Eddington approximation for radiative flux transfer, J. Atmos. Sci. 33, 2452-2459, 1976.

	#also see these lecture notes are pretty good
	#http://irina.eas.gatech.edu/EAS8803_SPRING2012/Lec20.pdf
	
	#w0=wbar_WDEL*(1.-cosb_CDEL**2)/(1.0-wbar_WDEL*cosb_CDEL**2)
	#cosbar=cosb_CDEL/(1.+cosb_CDEL)
	#dtau=dtau_DTDEL*(1.-wbar_WDEL*cosb_CDEL**2) 
	
	#sum up taus starting at the top, going to depth
	#tau = zeros((nlevel, nwno))
	#tau[1:,:]=numba_cumsum(dtau)

	#now define terms of Toon et al 1989 quadrature Table 1 
	#https://agupubs.onlinelibrary.wiley.com/doi/pdf/10.1029/JD094iD13p16287
	#see table of terms 
	sq3 = sqrt(3.)
	g1	= (sq3*0.5)*(2. - w0*(1.+cosbar))	#table 1
	g2	= (sq3*w0*0.5)*(1.-cosbar)		   #table 1
	g3	= 0.5*(1.-sq3*cosbar*ubar0)		  #table 1
	lamda = sqrt(g1**2 - g2**2)					 #eqn 21
	gama  = (g1-lamda)/g2							  #eqn 22
	
	# now calculate c_plus and c_minus (equation 23 and 24)
	g4 = 1.0 - g3
	denominator = lamda**2 - 1.0/ubar0**2.0

	#assert 0 in denomiator , ('Zero in C+, C- Toon+1989. Add catch.')

	#everything but the exponential 
	a_minus = F0PI*w0* (g4*(g1 + 1.0/ubar0) +g2*g3 ) / denominator
	a_plus  = F0PI*w0*(g3*(g1-1.0/ubar0) +g2*g4) / denominator

	#add in exponntial to get full eqn
	#_up is the terms evaluated at lower optical depths (higher altitudes)
	#_down is terms evaluated at higher optical depths (lower altitudes)
	x = exp(-tau[:-1,:]/ubar0)
	c_minus_up = a_minus*x #CMM1
	c_plus_up  = a_plus*x #CPM1
	x = exp(-tau[1:,:]/ubar0)
	c_minus_down = a_minus*x #CM
	c_plus_down  = a_plus*x #CP

	#calculate exponential terms needed for the tridiagonal rotated layered method
	exptrm = lamda*dtau
	#save from overflow 
	exptrm = slice_gt (exptrm, 35.0)

	exptrm_positive = exp(exptrm) #EP
	exptrm_minus = 1.0/exptrm_positive#exp(-exptrm) #EM


	#boundary conditions 
	b_top = 0.0										  
	b_surface = 0. + surf_reflect*ubar0*F0PI*exp(-tau[-1, :]/ubar0)

	#Now we need the terms for the tridiagonal rotated layered method
	A, B, C, D = setup_tri_diag(nlayer,nwno,  c_plus_up, c_minus_up, 
									c_plus_down, c_minus_down, b_top, b_surface, surf_reflect,
									g1, g2, g3, lamda, gama, dtau, 
									exptrm_positive,  exptrm_minus) 

	positive = zeros((nlayer, nwno))
	negative = zeros((nlayer, nwno))
	#========================= Start loop over wavelength =========================
	L = 2*nlayer
	for w in range(nwno):
		#coefficient of posive and negative exponential terms 
		X = tri_diag_solve(L, A[:,w], B[:,w], C[:,w], D[:,w])

		#unmix the coefficients
		positive[:,w] = X[::2] + X[1::2] 
		negative[:,w] = X[::2] - X[1::2]
	#========================= End loop over wavelength =========================

	#might have to add this in to avoid numerical problems later. 
	#if len(np.where(negative[:,w]/X[::2] < 1e-30)) >0 , print(negative[:,w],X[::2],negative[:,w]/X[::2])

	#evaluate the fluxes through the layers 
	#use the top optical depth expression to evaluate fp and fm 
	#at the top of each layer 
	flux_plus  = zeros((nlevel, nwno))
	flux_minus = zeros((nlevel, nwno))
	flux_plus[:-1,:]  = positive + gama*negative + c_plus_up #everything but the last row (botton of atmosphere)
	flux_minus[:-1,:] = positive*gama + negative + c_minus_up #everything but the last row (botton of atmosphere

	#use expression for bottom flux to get the flux_plus and flux_minus at last
	#bottom layer
	flux_plus[-1,:]  = positive[-1,:]*exptrm_positive[-1,:] + gama[-1,:]*negative[-1,:]*exptrm_minus[-1,:] + c_plus_down[-1,:]
	flux_minus[-1,:] = positive[-1,:]*exptrm_positive[-1,:]*gama[-1,:] + negative[-1,:]*exptrm_minus[-1,:] + c_minus_down[-1,:]

	#we have solved for fluxes directly and no further integration is needed 
	#ubar is absorbed into the definition of g1-g4
	#ubar=0.5: hemispheric constant 
	#ubar=sqrt(3): gauss quadrature 
	#other cases in meador & weaver JAS, 37, 630-643, 1980

	#now add direct flux term to the downwelling radiation, Liou 1982
	flux_minus = flux_minus + ubar0*F0PI*exp(-1.0*tau/ubar0)

	#now calculate the fluxes at the midpoints of the layers 
	#exptrm_positive_mdpt = exp(0.5*exptrm) #EP_mdpt
	#exptrm_minus_mdpt = exp(-0.5*exptrm) #EM_mdpt

	#tau_mdpt = tau[:-1] + 0.5*dtau #start from bottom up to define midpoints 
	#c_plus_mdpt = a_plus*exp(-tau_mdpt/ubar0)
	#c_minus_mdpt = a_minus*exp(-tau_mdpt/ubar0)
	
	#flux_plus_mdpt = positive*exptrm_positive_mdpt + gama*negative*exptrm_minus_mdpt + c_plus_mdpt
	#flux_minus_mdpt = positive*exptrm_positive_mdpt*gama + negative*exptrm_minus_mdpt + c_minus_mdpt

	#add direct flux to downwelling term 
	#flux_minus_mdpt = flux_minus_mdpt + ubar0*F0PI*exp(-1.0*tau_mdpt/ubar0)

	return flux_plus, flux_minus

@jit(nopython=True, cache=True)
def slice_eq(array, lim, value):
	"""Funciton to replace values with upper or lower limit
	"""
	for i in range(array.shape[0]):
		new = array[i,:] 
		new[where(new==lim)] = value
		array[i,:] = new	 
	return array

@jit(nopython=True, cache=True)
def slice_lt(array, lim):
	"""Funciton to replace values with upper or lower limit
	"""
	for i in range(array.shape[0]):
		new = array[i,:] 
		new[where(new<lim)] = lim
		array[i,:] = new	 
	return array

@jit(nopython=True, cache=True)
def slice_gt(array, lim):
	"""Funciton to replace values with upper or lower limit
	"""
	for i in range(array.shape[0]):
		new = array[i,:] 
		new[where(new>lim)] = lim
		array[i,:] = new	 
	return array

@jit(nopython=True, cache=True)
def numba_cumsum(mat):
	"""Function to compute cumsum along axis=0 to bypass numba not allowing kwargs in 
	cumsum 
	"""
	new_mat = zeros(mat.shape)
	for i in range(mat.shape[1]):
		new_mat[:,i] = cumsum(mat[:,i])
	return new_mat

@jit(nopython=True, cache=True)
def setup_tri_diag(nlayer,nwno ,c_plus_up, c_minus_up, 
	c_plus_down, c_minus_down, b_top, b_surface, surf_reflect,
	gama, dtau, exptrm_positive,  exptrm_minus):
	"""
	Before we can solve the tridiagonal matrix (See Toon+1989) section
	"SOLUTION OF THE TwO-STREAM EQUATIONS FOR MULTIPLE LAYERS", we 
	need to set up the coefficients. 

	Parameters
	----------
	nlayer : int 
		number of layers in the model 
	nwno : int 
		number of wavelength points
	c_plus_up : array 
		c-plus evaluated at the top of the atmosphere 
	c_minus_up : array 
		c_minus evaluated at the top of the atmosphere 
	c_plus_down : array 
		c_plus evaluated at the bottom of the atmosphere 
	c_minus_down : array 
		c_minus evaluated at the bottom of the atmosphere 
	b_top : array 
		The diffuse radiation into the model at the top of the atmosphere
	b_surface : array
		The diffuse radiation into the model at the bottom. Includes emission, reflection 
		of the unattenuated portion of the direct beam  
	surf_reflect : array 
		Surface reflectivity 
	g1 : array 
		table 1 toon et al 1989
	g2 : array 
		table 1 toon et al 1989
	g3 : array 
		table 1 toon et al 1989
	lamba : array 
		Eqn 21 toon et al 1989 
	gama : array 
		Eqn 22 toon et al 1989
	dtau : array 
		Opacity per layer
	exptrm_positive : array 
		Eqn 44, expoential terms needed for tridiagonal rotated layered, clipped at 35 
	exptrm_minus : array 
		Eqn 44, expoential terms needed for tridiagonal rotated layered, clipped at 35 


	Returns
	-------
	array 
		coefficient of the positive exponential term 
	
	"""
	L = 2 * nlayer

	#EQN 44 

	e1 = exptrm_positive + gama*exptrm_minus
	e2 = exptrm_positive - gama*exptrm_minus
	e3 = gama*exptrm_positive + exptrm_minus
	e4 = gama*exptrm_positive - exptrm_minus

	#now build terms 
	A = zeros((L,nwno)) 
	B = zeros((L,nwno )) 
	C = zeros((L,nwno )) 
	D = zeros((L,nwno )) 

	A[0,:] = 0.0
	B[0,:] = gama[0,:] + 1.0
	C[0,:] = gama[0,:] - 1.0
	D[0,:] = b_top - c_minus_up[0,:]

	#even terms, not including the last !CMM1 = UP
	A[1::2,:][:-1] = (e1[:-1,:]+e3[:-1,:]) * (gama[1:,:]-1.0) #always good
	B[1::2,:][:-1] = (e2[:-1,:]+e4[:-1,:]) * (gama[1:,:]-1.0)
	C[1::2,:][:-1] = 2.0 * (1.0-gama[1:,:]**2)			#always good 
	D[1::2,:][:-1] =((gama[1:,:]-1.0)*(c_plus_up[1:,:] - c_plus_down[:-1,:]) + 
							(1.0-gama[1:,:])*(c_minus_down[:-1,:] - c_minus_up[1:,:]))
	#import pickle as pk
	#pk.dump({'GAMA_1':(gama[1:,:]-1.0), 'CPM1':c_plus_up[1:,:] , 'CP':c_plus_down[:-1,:], '1_GAMA':(1.0-gama[1:,:]), 
	#	'CM':c_minus_down[:-1,:],'CMM1':c_minus_up[1:,:],'Deven':D[1::2,:][:-1]}, open('../testing_notebooks/GFLUX_even_D_terms.pk','wb'))
	
	#odd terms, not including the first 
	A[::2,:][1:] = 2.0*(1.0-gama[:-1,:]**2)
	B[::2,:][1:] = (e1[:-1,:]-e3[:-1,:]) * (gama[1:,:]+1.0)
	C[::2,:][1:] = (e1[:-1,:]+e3[:-1,:]) * (gama[1:,:]-1.0)
	D[::2,:][1:] = (e3[:-1,:]*(c_plus_up[1:,:] - c_plus_down[:-1,:]) + 
							e1[:-1,:]*(c_minus_down[:-1,:] - c_minus_up[1:,:]))

	#last term [L-1]
	A[-1,:] = e1[-1,:]-surf_reflect*e3[-1,:]
	B[-1,:] = e2[-1,:]-surf_reflect*e4[-1,:]
	C[-1,:] = 0.0
	D[-1,:] = b_surface-c_plus_down[-1,:] + surf_reflect*c_minus_down[-1,:]

	return A, B, C, D

@jit(nopython=True, cache=True)
def tri_diag_solve(l, a, b, c, d):
	"""
	Tridiagonal Matrix Algorithm solver, a b c d can be NumPy array type or Python list type.
	refer to this wiki_ and to this explanation_. 
	
	.. _wiki: http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm
	.. _explanation: http://www.cfd-online.com/Wiki/Tridiagonal_matrix_algorithm_-_TDMA_(Thomas_algorithm)
	
	A, B, C and D refer to: 

	.. math:: A(I)*X(I-1) + B(I)*X(I) + C(I)*X(I+1) = D(I)

	This solver returns X. 

	Parameters
	----------
	A : array or list 
	B : array or list 
	C : array or list 
	C : array or list 

	Returns
	-------
	array 
		Solution, x 
	"""
	AS, DS, CS, DS,XK = zeros(l), zeros(l), zeros(l), zeros(l), zeros(l) # copy arrays

	AS[-1] = a[-1]/b[-1]
	DS[-1] = d[-1]/b[-1]

	for i in range(l-2, -1, -1):
		x = 1.0 / (b[i] - c[i] * AS[i+1])
		AS[i] = a[i] * x
		DS[i] = (d[i]-c[i] * DS[i+1]) * x
	XK[0] = DS[0]
	for i in range(1,l):
		XK[i] = DS[i] - AS[i] * XK[i-1]
	return XK


@jit(nopython=True, cache=True)
def get_reflected_3d(nlevel, wno,nwno, numg,numt, dtau_3d, tau_3d, w0_3d, cosb_3d,gcos2_3d, ftau_cld_3d,ftau_ray_3d,
	dtau_og_3d, tau_og_3d, w0_og_3d, cosb_og_3d, 
	surf_reflect,ubar0, ubar1,cos_theta, F0PI,single_phase, multi_phase,
	frac_a, frac_b, frac_c, constant_back, constant_forward):
	"""
	Computes toon fluxes given tau and everything is 3 dimensional. This is the exact same function 
	as `get_flux_geom_1d` but is kept separately so we don't have to do unecessary indexing for 
	retrievals. 

	Parameters
	----------
	nlevel : int 
		Number of levels in the model 
	wno : array of float 
		Wave number grid in cm -1 
	nwno : int 
		Number of wave points
	numg : int 
		Number of Gauss angles 
	numt : int 
		Number of Chebyshev angles 
	dtau_3d : ndarray of float
		This is the opacity contained within each individual layer (defined at midpoints of "levels")
		WITHOUT D-Eddington Correction
		Dimensions=# layer by # wave
	tau_3d : ndarray of float
		This is the cumulative summed opacity 
		WITHOUT D-Eddington Correction
		Dimensions=# level by # wave		
	w0_3d : ndarray of float 
		This is the single scattering albedo, from scattering, clouds, raman, etc 
		WITHOUT D-Eddington Correction
		Dimensions=# layer by # wave
	cosb_3d : ndarray of float 
		This is the asymmetry factor 
		WITHOUT D-Eddington Correction
		Dimensions=# layer by # wave
	gcos2_3d : ndarray of float 
		Parameter that allows us to directly include Rayleigh scattering 
		= 0.5*tau_rayleigh/(tau_rayleigh + tau_cloud)
	ftau_cld_3d : ndarray of float 
		Fraction of cloud extinction to total 
		= tau_cloud/(tau_rayleigh + tau_cloud)
	ftau_ray_3d : ndarray of float 
		Fraction of rayleigh extinction to total 
		= tau_rayleigh/(tau_rayleigh + tau_cloud)
	dtau_og_3d : ndarray of float 
		This is the opacity contained within each individual layer (defined at midpoints of "levels")
		WITHOUT the delta eddington correction, if it was specified by user
		Dimensions=# layer by # wave
	tau_og_3d : ndarray of float
		This is the cumulative summed opacity 
		WITHOUT the delta eddington correction, if it was specified by user
		Dimensions=# level by # wave	
	w0_og_3d : ndarray of float 
		Same as w0 but WITHOUT the delta eddington correction, if it was specified by user	
	cosb_og_3d : ndarray of float 
		Same as cosbar buth WITHOUT the delta eddington correction, if it was specified by user
	surf_reflect : float 
		Surface reflectivity 
	ubar0 : ndarray of float 
		matrix of cosine of the incident angle from geometric.json
	ubar1 : ndarray of float 
		matrix of cosine of the observer angles
	cos_theta : float 
		Cosine of the phase angle of the planet 
	F0PI : array 
		Downward incident solar radiation
	single_phase : str 
		Single scattering phase function, default is the two-term henyey-greenstein phase function
 	multi_phase : str 
 		Multiple scattering phase function, defulat is N=2 Legendre polynomial
	frac_a : float 
		(Optional), If using the TTHG phase function. Must specify the functional form for fraction 
		of forward to back scattering (A + B * gcosb^C)
	frac_b : float 
		(Optional), If using the TTHG phase function. Must specify the functional form for fraction 
		of forward to back scattering (A + B * gcosb^C)
	frac_c : float 
		(Optional), If using the TTHG phase function. Must specify the functional form for fraction 
		of forward to back scattering (A + B * gcosb^C), Default is : 1 - gcosb^2
	constant_back : float 
		(Optional), If using the TTHG phase function. Must specify the assymetry of back scatterer. 
		Remember, the output of A & M code does not separate back and forward scattering.
	constant_forward : float 
		(Optional), If using the TTHG phase function. Must specify the assymetry of forward scatterer. 
		Remember, the output of A & M code does not separate back and forward scattering.
	Returns
	-------
	intensity at the top of the atmosphere for all the different ubar1 and ubar2 

	To Do
	-----
	- F0PI Solar flux shouldn't always be 1.. Follow up to make sure that this isn't a bad 
		  hardwiring to solar, despite "relative albedo"
	- take delta eddington option out of fluxes and move it all over to optics

	"""
	#what we want : intensity at the top as a function of all the different angles

	xint_at_top = zeros((numg, numt, nwno))

	nlayer = nlevel - 1 

	#now define terms of Toon et al 1989 quadrature Table 1 
	#https://agupubs.onlinelibrary.wiley.com/doi/pdf/10.1029/JD094iD13p16287
	#see table of terms 

	#terms not dependent on incident angle
	sq3 = sqrt(3.)

	#================ START CRAZE LOOP OVER ANGLE #================
	for ng in range(numg):
		for nt in range(numt):

			#get needed chunk for 3d inputs
			#should consider thinking of a better method for when doing 1d only
			cosb = cosb_3d[:,:,ng,nt]
			dtau = dtau_3d[:,:,ng,nt]
			tau = tau_3d[:,:,ng,nt]
			w0 = w0_3d[:,:,ng,nt]
			gcos2 = gcos2_3d[:,:,ng,nt]
			ftau_cld = ftau_cld_3d[:,:,ng,nt]
			ftau_ray = ftau_ray_3d[:,:,ng,nt]

			#uncorrected original values (in case user specified D-Eddington)
			#If they did not, this is the same thing as what is defined above 
			#These are used because HG single scattering phase function does get 
			#the forward and back scattering pretty accurately so delta-eddington
			#is only applied to the multiple scattering terms
			cosb_og = cosb_og_3d[:,:,ng,nt]
			dtau_og = dtau_og_3d[:,:,ng,nt]
			tau_og = tau_og_3d[:,:,ng,nt]
			w0_og = w0_og_3d[:,:,ng,nt]			

			g1	= (sq3*0.5)*(2. - w0*(1.+cosb))	#table 1
			g2	= (sq3*w0*0.5)*(1.-cosb)		   #table 1
			lamda = sqrt(g1**2 - g2**2)			  #eqn 21
			gama  = (g1-lamda)/g2					#eqn 22
			g3	= 0.5*(1.-sq3*cosb*ubar0[ng, nt])   #table 1 #ubar is now 100x 10 matrix.. 
	
			# now calculate c_plus and c_minus (equation 23 and 24)
			g4 = 1.0 - g3
			denominator = lamda**2 - 1.0/ubar0[ng, nt]**2.0

			#everything but the exponential 
			a_minus = F0PI*w0* (g4*(g1 + 1.0/ubar0[ng, nt]) +g2*g3 ) / denominator
			a_plus  = F0PI*w0*(g3*(g1-1.0/ubar0[ng, nt]) +g2*g4) / denominator

			#add in exponential to get full eqn
			#_up is the terms evaluated at lower optical depths (higher altitudes)
			#_down is terms evaluated at higher optical depths (lower altitudes)
			x = exp(-tau[:-1,:]/ubar0[ng, nt])
			c_minus_up = a_minus*x #CMM1
			c_plus_up  = a_plus*x #CPM1
			x = exp(-tau[1:,:]/ubar0[ng, nt])
			c_minus_down = a_minus*x #CM
			c_plus_down  = a_plus*x #CP

			#calculate exponential terms needed for the tridiagonal rotated layered method
			exptrm = lamda*dtau
			#save from overflow 
			exptrm = slice_gt (exptrm, 40.0) 

			exptrm_positive = exp(exptrm) #EP
			exptrm_minus = 1.0/exptrm_positive#exp(-exptrm) #EM


			#boundary conditions 
			b_top = 0.0										  
			b_surface = 0. + surf_reflect*ubar0[ng, nt]*F0PI*exp(-tau[-1, :]/ubar0[ng, nt])

			#Now we need the terms for the tridiagonal rotated layered method
			A, B, C, D = setup_tri_diag(nlayer,nwno,  c_plus_up, c_minus_up, 
									c_plus_down, c_minus_down, b_top, b_surface, surf_reflect,
									 gama, dtau, 
									exptrm_positive,  exptrm_minus) 

			positive = zeros((nlayer, nwno))
			negative = zeros((nlayer, nwno))
			#========================= Start loop over wavelength =========================
			L = 2*nlayer
			for w in range(nwno):
				#coefficient of posive and negative exponential terms 
				X = tri_diag_solve(L, A[:,w], B[:,w], C[:,w], D[:,w])

				#unmix the coefficients
				positive[:,w] = X[::2] + X[1::2] 
				negative[:,w] = X[::2] - X[1::2]
			#========================= End loop over wavelength =========================

			#use expression for bottom flux to get the flux_plus and flux_minus at last
			#bottom layer
			flux_zero  = positive[-1,:]*exptrm_positive[-1,:] + gama[-1,:]*negative[-1,:]*exptrm_minus[-1,:] + c_plus_down[-1,:]
			
			xint = zeros((nlevel,nwno))
			xint[-1,:] = flux_zero/pi

			################################ BEGIN OPTIONS FOR MULTIPLE SCATTERING####################

			#Legendre polynomials for the Phase function due to multiple scatterers 
			if multi_phase ==0:#'N=2':
				#ubar2 is defined to deal with the integration over the second moment of the 
				#intensity. It is FIT TO PURE RAYLEIGH LIMIT, ~(1/sqrt(3))^(1/2)
				#this is a decent assumption because our second order legendre polynomial 
				#is forced to be equal to the rayleigh phase function
				ubar2 = 0.767  # 
				multi_plus = (1.0+1.5*cosb*ubar1[ng,nt] #!was 3
								+ gcos2*(3.0*ubar2*ubar2*ubar1[ng,nt]*ubar1[ng,nt] - 1.0)/2.0)
				multi_minus = (1.-1.5*cosb*ubar1[ng,nt] 
								+ gcos2*(3.0*ubar2*ubar2*ubar1[ng,nt]*ubar1[ng,nt] - 1.0)/2.0)
			elif multi_phase ==1:#'N=1':
				multi_plus = 1.0+1.5*cosb*ubar1[ng,nt]	
				multi_minus = 1.-1.5*cosb*ubar1[ng,nt]
			################################ END OPTIONS FOR MULTIPLE SCATTERING####################

			G=w0*positive*(multi_plus+gama*multi_minus)
			H=w0*negative*(gama*multi_plus+multi_minus)
			A=w0*(multi_plus*c_plus_up+multi_minus*c_minus_up)

			G=G*0.5/pi
			H=H*0.5/pi
			A=A*0.5/pi

			################################ BEGIN OPTIONS FOR DIRECT SCATTERING####################
			#define f (fraction of forward to back scattering), 
			#g_forward (forward asymmetry), g_back (backward asym)
			#needed for everything except the OTHG
			if single_phase!=1: 
				g_forward = constant_forward*cosb_og
				g_back = -constant_back*cosb_og
				f = frac_a + frac_b*g_back**frac_c


			if single_phase==0:#'cahoy':
				#Phase function for single scattering albedo frum Solar beam
				#uses the Two term Henyey-Greenstein function with the additiona rayleigh component 
					  #first term of TTHG: forward scattering
				p_single=(f * (1-g_forward**2)
								/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
								#second term of TTHG: backward scattering
								+(1-f)*(1-g_back**2)
								/sqrt((1+(-cosb_og/2.)**2+2*(-cosb_og/2.)*cos_theta)**3)+
								#rayleigh phase function
								(gcos2))
			elif single_phase==1:#'OTHG':
				p_single=(1-cosb_og**2)/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
			elif single_phase==2:#'TTHG':
				#Phase function for single scattering albedo frum Solar beam
				#uses the Two term Henyey-Greenstein function with the additiona rayleigh component 
					  #first term of TTHG: forward scattering
				p_single=(f * (1-g_forward**2)
								/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
								#second term of TTHG: backward scattering
								+(1-f)*(1-g_back**2)
								/sqrt((1+(-cosb_og/2.)**2+2*(-cosb_og/2.)*cos_theta)**3))
			elif single_phase==3:#'TTHG_ray':
				#Phase function for single scattering albedo frum Solar beam
				#uses the Two term Henyey-Greenstein function with the additiona rayleigh component 
					  		#first term of TTHG: forward scattering
				p_single=(ftau_cld*(f * (1-g_forward**2)
												/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
												#second term of TTHG: backward scattering
												+(1-f)*(1-g_back**2)
												/sqrt((1+(-cosb_og/2.)**2+2*(-cosb_og/2.)*cos_theta)**3))+			
								#rayleigh phase function
								ftau_ray*(0.75*(1+cos_theta**2.0)))

			################################ END OPTIONS FOR DIRECT SCATTERING####################

			for i in range(nlayer-1,-1,-1):
				#direct beam
				#note when delta-eddington=off, then tau_single=tau, cosb_single=cosb, w0_single=w0, etc
				xint[i,:] =( xint[i+1,:]*exp(-dtau[i,:]/ubar1[ng,nt])
						#single scattering albedo from sun beam (from ubar0 to ubar1)
						+(w0_og[i,:]*F0PI/(4.*pi))*
						(p_single[i,:])*exp(-tau_og[i,:]/ubar0[ng,nt])*
						(1. - exp(-dtau_og[i,:]*(ubar0[ng,nt]+ubar1[ng,nt])/(ubar0[ng,nt]*ubar1[ng,nt])))*
						(ubar0[ng,nt]/(ubar0[ng,nt]+ubar1[ng,nt]))
						#three multiple scattering terms 
						+A[i,:]* (1. - exp(-dtau[i,:] *(ubar0[ng,nt]+1*ubar1[ng,nt])/(ubar0[ng,nt]*ubar1[ng,nt])))*
						(ubar0[ng,nt]/(ubar0[ng,nt]+1*ubar1[ng,nt]))
						+G[i,:]*(exp(exptrm[i,:]*1-dtau[i,:]/ubar1[ng,nt]) - 1.0)/(lamda[i,:]*1*ubar1[ng,nt] - 1.0)
						+H[i,:]*(1. - exp(-exptrm[i,:]*1-dtau[i,:]/ubar1[ng,nt]))/(lamda[i,:]*1*ubar1[ng,nt] + 1.0))
						#thermal

			xint_at_top[ng,nt,:] = xint[0,:]	
	return xint_at_top

@jit(nopython=True, cache=True)
def get_reflected_1d(nlevel, wno,nwno, numg,numt, dtau, tau, w0, cosb,gcos2, ftau_cld, ftau_ray,
	dtau_og, tau_og, w0_og, cosb_og, 
	surf_reflect,ubar0, ubar1,cos_theta, F0PI,single_phase, multi_phase,
	frac_a, frac_b, frac_c, constant_back, constant_forward):
	"""
	Computes toon fluxes given tau and everything is 1 dimensional. This is the exact same function 
	as `get_flux_geom_3d` but is kept separately so we don't have to do unecessary indexing for fast
	retrievals. 

	Parameters
	----------
	nlevel : int 
		Number of levels in the model 
	wno : array of float 
		Wave number grid in cm -1 
	nwno : int 
		Number of wave points
	numg : int 
		Number of Gauss angles 
	numt : int 
		Number of Chebyshev angles 
	DTAU : ndarray of float
		This is the opacity contained within each individual layer (defined at midpoints of "levels")
		WITHOUT D-Eddington Correction
		Dimensions=# layer by # wave
	TAU : ndarray of float
		This is the cumulative summed opacity 
		WITHOUT D-Eddington Correction
		Dimensions=# level by # wave		
	W0 : ndarray of float 
		This is the single scattering albedo, from scattering, clouds, raman, etc 
		WITHOUT D-Eddington Correction
		Dimensions=# layer by # wave
	COSB : ndarray of float 
		This is the asymmetry factor 
		WITHOUT D-Eddington Correction
		Dimensions=# layer by # wave
	GCOS2 : ndarray of float 
		Parameter that allows us to directly include Rayleigh scattering 
		= 0.5*tau_rayleigh/(tau_rayleigh + tau_cloud)
	ftau_cld : ndarray of float 
		Fraction of cloud extinction to total 
		= tau_cloud/(tau_rayleigh + tau_cloud)
	ftau_ray : ndarray of float 
		Fraction of rayleigh extinction to total 
		= tau_rayleigh/(tau_rayleigh + tau_cloud)
	dtau_og : ndarray of float 
		This is the opacity contained within each individual layer (defined at midpoints of "levels")
		WITHOUT the delta eddington correction, if it was specified by user
		Dimensions=# layer by # wave
	tau_og : ndarray of float
		This is the cumulative summed opacity 
		WITHOUT the delta eddington correction, if it was specified by user
		Dimensions=# level by # wave	
	w0_og : ndarray of float 
		Same as w0 but WITHOUT the delta eddington correction, if it was specified by user	
	cosb_og : ndarray of float 
		Same as cosbar buth WITHOUT the delta eddington correction, if it was specified by user
	surf_reflect : float 
		Surface reflectivity 
	ubar0 : ndarray of float 
		matrix of cosine of the incident angle from geometric.json
	ubar1 : ndarray of float 
		matrix of cosine of the observer angles
	cos_theta : float 
		Cosine of the phase angle of the planet 
	F0PI : array 
		Downward incident solar radiation
	single_phase : str 
		Single scattering phase function, default is the two-term henyey-greenstein phase function
 	multi_phase : str 
 		Multiple scattering phase function, defulat is N=2 Legendre polynomial 
	frac_a : float 
		(Optional), If using the TTHG phase function. Must specify the functional form for fraction 
		of forward to back scattering (A + B * gcosb^C)
	frac_b : float 
		(Optional), If using the TTHG phase function. Must specify the functional form for fraction 
		of forward to back scattering (A + B * gcosb^C)
	frac_c : float 
		(Optional), If using the TTHG phase function. Must specify the functional form for fraction 
		of forward to back scattering (A + B * gcosb^C), Default is : 1 - gcosb^2
	constant_back : float 
		(Optional), If using the TTHG phase function. Must specify the assymetry of back scatterer. 
		Remember, the output of A & M code does not separate back and forward scattering.
	constant_forward : float 
		(Optional), If using the TTHG phase function. Must specify the assymetry of forward scatterer. 
		Remember, the output of A & M code does not separate back and forward scattering.
	Returns
	-------
	intensity at the top of the atmosphere for all the different ubar1 and ubar2 

	To Do
	-----
	- F0PI Solar flux shouldn't always be 1.. Follow up to make sure that this isn't a bad 
		  hardwiring to solar, despite "relative albedo"
	"""
	#what we want : intensity at the top as a function of all the different angles

	xint_at_top = zeros((numg, numt, nwno))

	nlayer = nlevel - 1 

	#now define terms of Toon et al 1989 quadrature Table 1 
	#https://agupubs.onlinelibrary.wiley.com/doi/pdf/10.1029/JD094iD13p16287
	#see table of terms 

	#terms not dependent on incident angle
	sq3 = sqrt(3.)
	g1	= (sq3*0.5)*(2. - w0*(1.+cosb))	#table 1
	g2	= (sq3*w0*0.5)*(1.-cosb)		#table 1
	lamda = sqrt(g1**2 - g2**2)			#eqn 21
	gama  = (g1-lamda)/g2				#eqn 22

	#================ START CRAZE LOOP OVER ANGLE #================
	for ng in range(numg):
		for nt in range(numt):

			g3	= 0.5*(1.-sq3*cosb*ubar0[ng, nt])   #table 1 #ubar has dimensions [gauss angles by tchebyshev angles ]
	
			# now calculate c_plus and c_minus (equation 23 and 24 toon)
			g4 = 1.0 - g3
			denominator = lamda**2 - 1.0/ubar0[ng, nt]**2.0

			#everything but the exponential 
			a_minus = F0PI*w0* (g4*(g1 + 1.0/ubar0[ng, nt]) +g2*g3 ) / denominator
			a_plus  = F0PI*w0*(g3*(g1-1.0/ubar0[ng, nt]) +g2*g4) / denominator

			#add in exponential to get full eqn
			#_up is the terms evaluated at lower optical depths (higher altitudes)
			#_down is terms evaluated at higher optical depths (lower altitudes)
			x = exp(-tau[:-1,:]/ubar0[ng, nt])
			c_minus_up = a_minus*x #CMM1
			c_plus_up  = a_plus*x #CPM1
			x = exp(-tau[1:,:]/ubar0[ng, nt])
			c_minus_down = a_minus*x #CM
			c_plus_down  = a_plus*x #CP

			#calculate exponential terms needed for the tridiagonal rotated layered method
			exptrm = lamda*dtau
			#save from overflow 
			exptrm = slice_gt (exptrm, 35.0) 

			exptrm_positive = exp(exptrm) #EP
			exptrm_minus = 1.0/exptrm_positive#exp(-exptrm) #EM


			#boundary conditions 
			b_top = 0.0										  
			b_surface = 0. + surf_reflect*ubar0[ng, nt]*F0PI*exp(-tau[-1, :]/ubar0[ng, nt])

			#Now we need the terms for the tridiagonal rotated layered method
			A, B, C, D = setup_tri_diag(nlayer,nwno,  c_plus_up, c_minus_up, 
									c_plus_down, c_minus_down, b_top, b_surface, surf_reflect,
									gama, dtau, 
									exptrm_positive,  exptrm_minus) 

			positive = zeros((nlayer, nwno))
			negative = zeros((nlayer, nwno))
			#========================= Start loop over wavelength =========================
			L = 2*nlayer
			for w in range(nwno):
				#coefficient of posive and negative exponential terms 
				X = tri_diag_solve(L, A[:,w], B[:,w], C[:,w], D[:,w])

				#unmix the coefficients
				positive[:,w] = X[::2] + X[1::2] 
				negative[:,w] = X[::2] - X[1::2]
			#========================= End loop over wavelength =========================

			#use expression for bottom flux to get the flux_plus and flux_minus at last
			#bottom layer
			flux_zero  = positive[-1,:]*exptrm_positive[-1,:] + gama[-1,:]*negative[-1,:]*exptrm_minus[-1,:] + c_plus_down[-1,:]
			
			xint = zeros((nlevel,nwno))
			xint[-1,:] = flux_zero/pi

			################################ BEGIN OPTIONS FOR MULTIPLE SCATTERING####################

			#Legendre polynomials for the Phase function due to multiple scatterers 
			if multi_phase ==0:#'N=2':
				#ubar2 is defined to deal with the integration over the second moment of the 
				#intensity. It is FIT TO PURE RAYLEIGH LIMIT, ~(1/sqrt(3))^(1/2)
				#this is a decent assumption because our second order legendre polynomial 
				#is forced to be equal to the rayleigh phase function
				ubar2 = 0.767  # 
				multi_plus = (1.0+1.5*cosb*ubar1[ng,nt] #!was 3
								+ gcos2*(3.0*ubar2*ubar2*ubar1[ng,nt]*ubar1[ng,nt] - 1.0)/2.0)
				multi_minus = (1.-1.5*cosb*ubar1[ng,nt] 
								+ gcos2*(3.0*ubar2*ubar2*ubar1[ng,nt]*ubar1[ng,nt] - 1.0)/2.0)
			elif multi_phase ==1:#'N=1':
				multi_plus = 1.0+1.5*cosb*ubar1[ng,nt]	
				multi_minus = 1.-1.5*cosb*ubar1[ng,nt]
			################################ END OPTIONS FOR MULTIPLE SCATTERING####################

			G=w0*positive*(multi_plus+gama*multi_minus)
			H=w0*negative*(gama*multi_plus+multi_minus)
			A=w0*(multi_plus*c_plus_up+multi_minus*c_minus_up)

			G=G*0.5/pi
			H=H*0.5/pi
			A=A*0.5/pi

			################################ BEGIN OPTIONS FOR DIRECT SCATTERING####################
			#define f (fraction of forward to back scattering), 
			#g_forward (forward asymmetry), g_back (backward asym)
			#needed for everything except the OTHG
			if single_phase!=1: 
				g_forward = constant_forward*cosb_og
				g_back = -constant_back*cosb_og
				f = frac_a + frac_b*g_back**frac_c


			if single_phase==0:#'cahoy':
				#Phase function for single scattering albedo frum Solar beam
				#uses the Two term Henyey-Greenstein function with the additiona rayleigh component 
					  #first term of TTHG: forward scattering
				p_single=(f * (1-g_forward**2)
								/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
								#second term of TTHG: backward scattering
								+(1-f)*(1-g_back**2)
								/sqrt((1+(-cosb_og/2.)**2+2*(-cosb_og/2.)*cos_theta)**3)+
								#rayleigh phase function
								(gcos2))
			elif single_phase==1:#'OTHG':
				p_single=(1-cosb_og**2)/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
			elif single_phase==2:#'TTHG':
				#Phase function for single scattering albedo frum Solar beam
				#uses the Two term Henyey-Greenstein function with the additiona rayleigh component 
					  #first term of TTHG: forward scattering
				p_single=(f * (1-g_forward**2)
								/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
								#second term of TTHG: backward scattering
								+(1-f)*(1-g_back**2)
								/sqrt((1+(-cosb_og/2.)**2+2*(-cosb_og/2.)*cos_theta)**3))
			elif single_phase==3:#'TTHG_ray':
				#Phase function for single scattering albedo frum Solar beam
				#uses the Two term Henyey-Greenstein function with the additiona rayleigh component 
					  		#first term of TTHG: forward scattering
				p_single=(ftau_cld*(f * (1-g_forward**2)
												/sqrt((1+cosb_og**2+2*cosb_og*cos_theta)**3) 
												#second term of TTHG: backward scattering
												+(1-f)*(1-g_back**2)
												/sqrt((1+(-cosb_og/2.)**2+2*(-cosb_og/2.)*cos_theta)**3))+			
								#rayleigh phase function
								ftau_ray*(0.75*(1+cos_theta**2.0)))

			################################ END OPTIONS FOR DIRECT SCATTERING####################

			for i in range(nlayer-1,-1,-1):
						#direct beam
				xint[i,:] =( xint[i+1,:]*exp(-dtau[i,:]/ubar1[ng,nt]) 
						#single scattering albedo from sun beam (from ubar0 to ubar1)
						+(w0_og[i,:]*F0PI/(4.*pi))*
						(p_single[i,:])*exp(-tau_og[i,:]/ubar0[ng,nt])*
						(1. - exp(-dtau_og[i,:]*(ubar0[ng,nt]+ubar1[ng,nt])/(ubar0[ng,nt]*ubar1[ng,nt])))*
						(ubar0[ng,nt]/(ubar0[ng,nt]+ubar1[ng,nt]))
						#multiple scattering terms p_single
						+A[i,:]*(1. - exp(-dtau[i,:] *(ubar0[ng,nt]+1*ubar1[ng,nt])/(ubar0[ng,nt]*ubar1[ng,nt])))*
						(ubar0[ng,nt]/(ubar0[ng,nt]+1*ubar1[ng,nt]))
						+G[i,:]*(exp(exptrm[i,:]*1-dtau[i,:]/ubar1[ng,nt]) - 1.0)/(lamda[i,:]*1*ubar1[ng,nt] - 1.0)
						+H[i,:]*(1. - exp(-exptrm[i,:]*1-dtau[i,:]/ubar1[ng,nt]))/(lamda[i,:]*1*ubar1[ng,nt] + 1.0))
			xint_at_top[ng,nt,:] = xint[0,:]	
	return xint_at_top

@jit(nopython=True, cache=True)
def blackbody(t,w):
	"""
	Blackbody flux in cgs units in per unit wavelength

	Parameters
	----------
	t : array,float
		Temperature (K)
	w : array, float
		Wavelength (cgs)
	
	Returns
	-------
	ndarray with shape ntemp x numwave
	"""
	h = 6.62607004e-27 # erg s 
	c = 2.99792458e+10 # cm/s
	k = 1.38064852e-16 #erg / K

	return ((2.0*h*c**2.0)/(w**5.0))*(1.0/(exp((h*c)/outer(t, w*k)) - 1.0))

@jit(nopython=True, cache=True)
def get_thermal_1d(nlevel, wno,nwno, numg,numt,tlevel, dtau, w0,cosb,plevel, ubar1):
	
	nlayer = nlevel -1 #nlayers 

	mu1 = 0.5 #from Table 1 Toon  
	twopi = pi+pi

	#get matrix of blackbodies 
	all_b = blackbody(tlevel, 1/wno) #returns nlevel by nwave	
	b0 = all_b[0:-1,:]
	b1 = (all_b[1:,:] - b0) / dtau # eqn 26 toon 89

	#hemispheric mean parameters from Tabe 1 toon 
	alpha = sqrt( (1.-w0) / (1.-w0*cosb) )
	lamda = alpha*(1.-w0*cosb)/mu1 #eqn 21 toon
	gama = (1.-alpha)/(1.+alpha) #eqn 22 toon
	g1_plus_g2 = mu1/(1.-w0*cosb) #effectively 1/(gamma1 + gamma2) .. second half of eqn.27


	#same as with reflected light, compute c_plus and c_minus 
	#these are eqns 27a & b in Toon89
	#_ups are evaluated at lower optical depth, TOA
	#_dows are evaluated at higher optical depth, bottom of atmosphere
	c_plus_up = b0 + b1* g1_plus_g2
	c_minus_up = b0 - b1* g1_plus_g2

	c_plus_down = b0 + b1 * dtau + b1 * g1_plus_g2 
	c_minus_down = b0 + b1 * dtau - b1 * g1_plus_g2

	#calculate exponential terms needed for the tridiagonal rotated layered method
	exptrm = lamda*dtau
	#save from overflow 
	exptrm = slice_gt (exptrm, 35.0) 

	exptrm_positive = exp(exptrm) 
	exptrm_minus = 1.0/exptrm_positive#exp(-exptrm) 

	tau_top = dtau[0,:]*plevel[0]/(plevel[1]-plevel[0])
	b_top = (1.0 - exp(-tau_top / mu1 )) * all_b[0,:] 
	b_surface = all_b[-1,:] + b1[-1,:]*mu1
	surf_reflect = 0.

	#Now we need the terms for the tridiagonal rotated layered method
	A, B, C, D = setup_tri_diag(nlayer,nwno,  c_plus_up, c_minus_up, 
							c_plus_down, c_minus_down, b_top, b_surface, surf_reflect,
							gama, dtau, 
							exptrm_positive,  exptrm_minus) 
	positive = zeros((nlayer, nwno))
	negative = zeros((nlayer, nwno))
	#========================= Start loop over wavelength =========================
	L = nlayer+nlayer
	for w in range(nwno):
		#coefficient of posive and negative exponential terms 
		X = tri_diag_solve(L, A[:,w], B[:,w], C[:,w], D[:,w])

		#unmix the coefficients
		positive[:,w] = X[::2] + X[1::2] 
		negative[:,w] = X[::2] - X[1::2]

	f_up = pi*(positive * exptrm_positive + gama * negative * exptrm_minus + c_plus_up)

	#calculate everyting from Table 3 toon
	alphax = ((1.0-w0)/(1.0-w0*cosb))**0.5
	G = twopi*w0*positive*(1.0+cosb*alphax)/(1.0+alphax)
	H = twopi*w0*negative*(1.0-cosb*alphax)/(1.0+alphax)
	J = twopi*w0*positive*(1.0-cosb*alphax)/(1.0+alphax)
	K = twopi*w0*negative*(1.0+cosb*alphax)/(1.0+alphax)
	alpha1 = twopi*(b0+ b1*(mu1*w0*cosb/(1.0-w0*cosb)))
	alpha2 = twopi*b1
	sigma1 = twopi*(b0- b1*(mu1*w0*cosb/(1.0-w0*cosb)))
	sigma2 = twopi*b1

	flux_minus = zeros((nlevel,nwno))
	flux_plus = zeros((nlevel,nwno))
	flux_minus_mdpt = zeros((nlevel,nwno))
	flux_plus_mdpt = zeros((nlevel,nwno))

	exptrm_positive_mdpt = exp(0.5*exptrm) 
	exptrm_minus_mdpt = 1/exptrm_positive_mdpt 

	#================ START CRAZE LOOP OVER ANGLE #================
	flux_at_top = zeros((numg, numt, nwno))
	flux_down = zeros((numg, numt, nwno))

	#work through building eqn 55 in toon (tons of bookeeping exponentials)
	for ng in range(numg):
		for nt in range(numt): 
			flux_plus[-1,:] = twopi * (b_surface + b1[-1,:] * ubar1[ng,nt])
			flux_minus[0,:] = twopi * (1 - exp(-tau_top / ubar1[ng,nt])) * all_b[0,:]
			
			exptrm_angle = exp( - dtau / ubar1[ng,nt])
			exptrm_angle_mdpt = exp( -0.5 * dtau / ubar1[ng,nt]) 

			for itop in range(nlayer):

				#disbanning this for now because we dont need it in the thermal emission code
				#flux_minus[itop+1,:]=(flux_minus[itop,:]*exptrm_angle[itop,:]+
				#	                 (J[itop,:]/(lamda[itop,:]*ubar1[ng,nt]+1.0))*(exptrm_positive[itop,:]-exptrm_angle[itop,:])+
				#	                 (K[itop,:]/(lamda[itop,:]*ubar1[ng,nt]-1.0))*(exptrm_angle[itop,:]-exptrm_minus[itop,:])+
				#	                 sigma1[itop,:]*(1.-exptrm_angle[itop,:])+
				#	                 sigma2[itop,:]*(ubar1[ng,nt]*exptrm_angle[itop,:]+dtau[itop,:]-ubar1[ng,nt]) )

				#flux_minus_mdpt[itop,:]=(flux_minus[itop,:]*exptrm_angle_mdpt[itop,:]+
				#                        (J[itop,:]/(lamda[itop,:]*ubar1[ng,nt]+1.0))*(exptrm_positive_mdpt[itop,:]-exptrm_angle_mdpt[itop,:])+
				#                        (K[itop,:]/(-lamda[itop,:]*ubar1[ng,nt]+1.0))*(exptrm_minus_mdpt[itop,:]-exptrm_angle_mdpt[itop,:])+
				#                        sigma1[itop,:]*(1.-exptrm_angle_mdpt[itop,:])+
				#                        sigma2[itop,:]*(ubar1[ng,nt]*exptrm_angle_mdpt[itop,:]+0.5*dtau[itop,:]-ubar1[ng,nt]))

				ibot=nlayer-1-itop

				flux_plus[ibot,:]=(flux_plus[ibot+1,:]*exptrm_angle[ibot,:]+
				                  (G[ibot,:]/(lamda[ibot,:]*ubar1[ng,nt]-1.0))*(exptrm_positive[ibot,:]*exptrm_angle[ibot,:]-1.0)+
				                  (H[ibot,:]/(lamda[ibot,:]*ubar1[ng,nt]+1.0))*(1.0-exptrm_minus[ibot,:] * exptrm_angle[ibot,:])+
				                  alpha1[ibot,:]*(1.-exptrm_angle[ibot,:])+
				                  alpha2[ibot,:]*(ubar1[ng,nt]-(dtau[ibot,:]+ubar1[ng,nt])*exptrm_angle[ibot,:]) )

				flux_plus_mdpt[ibot,:]=(flux_plus[ibot+1,:]*exptrm_angle_mdpt[ibot,:]+
				                       (G[ibot,:]/(lamda[ibot,:]*ubar1[ng,nt]-1.0))*(exptrm_positive[ibot,:]*exptrm_angle_mdpt[ibot,:]-exptrm_positive_mdpt[ibot,:])-
				                       (H[ibot,:]/(lamda[ibot,:]*ubar1[ng,nt]+1.0))*(exptrm_minus[ibot,:]*exptrm_angle_mdpt[ibot,:]-exptrm_minus_mdpt[ibot,:])+
				                       alpha1[ibot,:]*(1.-exptrm_angle_mdpt[ibot,:])+
				                       alpha2[ibot,:]*(ubar1[ng,nt]+0.5*dtau[ibot,:]-(dtau[ibot,:]+ubar1[ng,nt])*exptrm_angle_mdpt[ibot,:])  )

		flux_at_top[ng,nt,:] = flux_plus_mdpt[0,:] #nlevel by nwno
		#flux_down[ng,nt,:] = flux_minus_mdpt[0,:] #nlevel by nwno, Dont really need to compute this for now
		
		#for testing purposes
		#import pickle as pk
		#pk.dump([flux_at_top, flux_down, flux_minus, flux_minus_mdpt, flux_plus, flux_plus_mdpt], open('/Users/natashabatalha/Documents/PICASO/test/thermal/debugthermal.pk','wb'))

	return flux_at_top #, flux_down# numg x numt x nwno



from bokeh.plotting import figure, output_file, show
from bokeh.palettes import Colorblind8
import numpy as np
import pandas as pd
from bokeh.layouts import column,row
import numpy as np
from bokeh.palettes import magma as colfun1
from bokeh.palettes import viridis as colfun2
from bokeh.palettes import gray as colfun3
from bokeh.palettes import Spectral11
from bokeh.models import HoverTool
from bokeh.models import LinearColorMapper, LogTicker,BasicTicker, ColorBar,LogColorMapper,Legend
import os 
from numba import jit
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import matplotlib.pyplot as plt
import holoviews as hv
from holoviews.operation.datashader import datashade
hv.extension('bokeh')

def mixing_ratio(full_output,**kwargs):
	"""Returns plot of mixing ratios 

	Parameters
	----------
	full_output : class
		picaso.atmsetup.ATMSETUP
	**kwargs : dict 
		Any key word argument for bokeh.figure() 
	"""
	#set plot defaults
	molecules = full_output['weights'].keys()
	pressure = full_output['layer']['pressure']

	kwargs['plot_height'] = kwargs.get('plot_height',300)
	kwargs['plot_width'] = kwargs.get('plot_width',400)
	kwargs['title'] = kwargs.get('title','Mixing Ratios')
	kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure(Bars)')
	kwargs['x_axis_label'] = kwargs.get('x_axis_label','Mixing Ratio(v/v)')
	kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
	kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')	
	kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure),np.min(pressure)])
	kwargs['x_range'] = kwargs.get('x_range',[1e-20, 1e2])


	fig = figure(**kwargs)
	cols = colfun1(len(molecules))
	legend_it=[]	
	for mol , c in zip(molecules,cols):
		ind = np.where(mol==np.array(molecules))[0][0]
		f = fig.line(full_output['layer']['mixingratios'][mol],pressure, color=c, line_width=3,
					muted_color=c, muted_alpha=0.2)
		legend_it.append((mol, [f]))

	legend = Legend(items=legend_it, location=(0, -20))
	legend.click_policy="mute"
	fig.add_layout(legend, 'left')  

	return fig
	
def pt(full_output,**kwargs):
	"""Returns plot of pressure temperature profile

	Parameters
	----------
	full_output : class
		picaso.atmsetup.ATMSETUP
	**kwargs : dict 
		Any key word argument for bokeh.figure() 
	"""
	#set plot defaults
	pressure = full_output['layer']['pressure']
	temperature = full_output['layer']['temperature']

	kwargs['plot_height'] = kwargs.get('plot_height',300)
	kwargs['plot_width'] = kwargs.get('plot_width',400)
	kwargs['title'] = kwargs.get('title','Pressure-Temperature Profile')
	kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure(Bars)')
	kwargs['x_axis_label'] = kwargs.get('x_axis_label','Temperature (K)')
	kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
	kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')	
	kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure),np.min(pressure)])

	fig = figure(**kwargs)

	f = fig.line(temperature,pressure,line_width=3) 
	plot_format(fig)
	return fig

def spectrum(wno, alb,legend=None, **kwargs):
	"""Plot formated albedo spectrum

	Parameters
	----------
	wno : float array, list of arrays
		wavenumber 
	alb : float array, list of arrays 
		albedo 
	legend : list of str 
		legends for plotting 
	**kwargs : dict 	
		Any key word argument for bokeh.plotting.figure()

	Returns
	-------
	bokeh plot
	"""
	kwargs['plot_height'] = kwargs.get('plot_height',345)
	kwargs['plot_width'] = kwargs.get('plot_width',1000)
	kwargs['y_axis_label'] = kwargs.get('y_axis_label','Albedo')
	kwargs['x_axis_label'] = kwargs.get('x_axis_label','Wavelength [μm]')
	kwargs['y_range'] = kwargs.get('y_range',[0,1.2])
	kwargs['x_range'] = kwargs.get('x_range',[0.3,1])

	fig = figure(**kwargs)

	if isinstance(wno, list):
		if legend == None: legend=[None]*len(wno) 
		for w, a,i,l in zip(wno, alb, range(len(wno)), legend):
			fig.line(1e4/w, a, legend=l, color=Colorblind8[np.mod(i, len(Colorblind8))], line_width=3)
	else: 
		fig.line(1e4/wno, alb, legend=legend, color=Colorblind8[0], line_width=3)

	plot_format(fig)
	return fig


def photon_attenuation(full_output, at_tau=0.5,**kwargs):
	"""
	Plot breakdown of gas opacity, cloud opacity, 
	Rayleigh scattering opacity at a specified pressure level. 
	
	Parameters
	----------
	full_output : class 
		picaso.atmsetup.ATMSETUP
	at_tau : float 
		Opacity at which to plot the cumulative opacity. 
		Default =1 bar. 
	**kwargs : dict 
		Any key word argument for bokeh.plotting.figure()

	Returns
	-------
	bokeh plot
	"""
	wave = 1e4/full_output['wavenumber']

	dtaugas = full_output['taugas']
	dtaucld = full_output['taucld']*full_output['layer']['cloud']['w0']
	dtauray = full_output['tauray']
	shape = dtauray.shape
	taugas = np.zeros((shape[0]+1, shape[1]))
	taucld = np.zeros((shape[0]+1, shape[1]))
	tauray = np.zeros((shape[0]+1, shape[1]))

	#comptue cumulative opacity
	taugas[1:,:]=numba_cumsum(dtaugas)
	taucld[1:,:]=numba_cumsum(dtaucld)
	tauray[1:,:]=numba_cumsum(dtauray)


	pressure = full_output['level']['pressure']

	at_pressures = np.zeros(shape[1]) #pressure for each wave point

	ind_gas = find_nearest(taugas, at_tau)
	ind_cld = find_nearest(taucld, at_tau)
	ind_ray = find_nearest(tauray, at_tau)

	if (len(taucld[taucld == 0]) == taucld.shape[0]*taucld.shape[1]) : 
		ind_cld = ind_cld*0 + shape[0]

	at_pressures_gas = np.zeros(shape[1])
	at_pressures_cld = np.zeros(shape[1])
	at_pressures_ray = np.zeros(shape[1])

	for i in range(shape[1]):
		at_pressures_gas[i] = pressure[ind_gas[i]]
		at_pressures_cld[i] = pressure[ind_cld[i]]
		at_pressures_ray[i] = pressure[ind_ray[i]]

	kwargs['plot_height'] = kwargs.get('plot_height',300)
	kwargs['plot_width'] = kwargs.get('plot_width',1000)
	kwargs['title'] = kwargs.get('title','Pressure at 𝞽 =' +str(at_tau))
	kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure(Bars)')
	kwargs['x_axis_label'] = kwargs.get('x_axis_label','Wavelength [μm]')
	kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
	kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure),1e-2])

	fig = figure(**kwargs)

	legend_it = []

	f = fig.line(wave,at_pressures_gas,line_width=3, color=Colorblind8[0]) 
	legend_it.append(('Gas Opacity', [f]))
	f = fig.line(wave,at_pressures_cld,line_width=3, color=Colorblind8[3]) 
	legend_it.append(('Cloud Opacity', [f]))
	f = fig.line(wave,at_pressures_ray,line_width=3,color=Colorblind8[6]) 
	legend_it.append(('Rayleigh Opacity', [f]))

	legend = Legend(items=legend_it, location=(0, -20))
	legend.click_policy="mute"
	fig.add_layout(legend, 'right')   

	#finally add color sections 
	gas_dominate_ind = np.where((at_pressures_gas<at_pressures_cld) & (at_pressures_gas<at_pressures_ray))[0]
	cld_dominate_ind = np.where((at_pressures_cld<at_pressures_gas) & (at_pressures_cld<at_pressures_ray))[0]
	ray_dominate_ind = np.where((at_pressures_ray<at_pressures_cld) & (at_pressures_ray<at_pressures_gas))[0]

	gas_dominate = np.zeros(shape[1]) + 1e-8
	cld_dominate = np.zeros(shape[1]) + 1e-8
	ray_dominate = np.zeros(shape[1])+ 1e-8

	gas_dominate[gas_dominate_ind] = at_pressures_gas[gas_dominate_ind]
	cld_dominate[cld_dominate_ind] = at_pressures_cld[cld_dominate_ind]
	ray_dominate[ray_dominate_ind] = at_pressures_ray[ray_dominate_ind]

	if len(gas_dominate) > 0  :
		band_x = np.append(np.array(wave), np.array(wave[::-1]))
		band_y = np.append(np.array(gas_dominate), np.array(gas_dominate)[::-1]*0+1e-8)
		fig.patch(band_x,band_y, color=Colorblind8[0], alpha=0.3)
	if len(cld_dominate) > 0  :
		band_x = np.append(np.array(wave), np.array(wave[::-1]))
		band_y = np.append(np.array(cld_dominate), np.array(cld_dominate)[::-1]*0+1e-8)
		fig.patch(band_x,band_y, color=Colorblind8[3], alpha=0.3)
	if len(ray_dominate) > 0  :
		band_x = np.append(np.array(wave), np.array(wave[::-1]))
		band_y = np.append(np.array(ray_dominate), np.array(ray_dominate)[::-1]*0+1e-8)
		fig.patch(band_x,band_y, color=Colorblind8[6], alpha=0.3)

	plot_format(fig)
	return fig #,wave,at_pressures_gas,at_pressures_cld,at_pressures_ray

def plot_format(df):
	"""Function to reformat plots"""
	df.xaxis.axis_label_text_font='times'
	df.yaxis.axis_label_text_font='times'
	df.xaxis.major_label_text_font_size='14pt'
	df.yaxis.major_label_text_font_size='14pt'
	df.xaxis.axis_label_text_font_size='14pt'
	df.yaxis.axis_label_text_font_size='14pt'
	df.xaxis.major_label_text_font='times'
	df.yaxis.major_label_text_font='times'
	df.xaxis.axis_label_text_font_style = 'bold'
	df.yaxis.axis_label_text_font_style = 'bold'


def plot_cld_input(nwno, nlayer, filename=None,df=None,**pd_kwargs):
	"""
	This function was created to investigate CLD input file for PICASO. 

	The plot itselfs creates maps of the wavelength dependent single scattering albedo 
	and cloud opacity and assymetry parameter as a function of altitude. 


	Parameters
	----------
	nwno : int 
		Number of wavenumber points. For runs from Ackerman & Marley, this will always be 196. 
	nlayer : int 
		Should be one less than the number of levels in your pressure temperature grid. Cloud 
		opacity is assigned for slabs. 
	file : str 
		(Optional)Path to cloud input file
	df : str 
		(Optional)Dataframe of cloud input file
	pd_kwargs : kwargs
		Pandas key word arguments for `pandas.read_csv`

	Returns
	-------
	Three bokeh plots with the single scattering, optical depth, and assymetry maps
	"""
	cols = colfun1(200)
	color_mapper = LinearColorMapper(palette=cols, low=0, high=1)

	if not isinstance(filename,type(None)):
		dat01 = pd.read_csv(filename, **pd_kwargs)
	elif not isinstance(df,type(None)):
		dat01=df

	#PLOT W0
	scat01 = np.flip(np.reshape(dat01['w0'].values,(nlayer,nwno)),0)
	xr, yr = scat01.shape
	f01a = figure(x_range=[0, yr], y_range=[0,xr],
						   x_axis_label='Wavenumber Grid', y_axis_label='Pressure Grid, TOA ->',
						   title="Single Scattering Albedo",
						  plot_width=300, plot_height=300)


	f01a.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

	color_bar = ColorBar(color_mapper=color_mapper, #ticker=LogTicker(),
					   label_standoff=12, border_line_color=None, location=(0,0))

	f01a.add_layout(color_bar, 'left')


	#PLOT OPD
	scat01 = np.flip(np.reshape(dat01['opd'].values,(nlayer,nwno)),0)

	xr, yr = scat01.shape
	cols = colfun2(200)[::-1]
	color_mapper = LogColorMapper(palette=cols, low=1e-3, high=10)


	f01 = figure(x_range=[0, yr], y_range=[0,xr],
						   x_axis_label='Wavenumber Grid', y_axis_label='Pressure Grid, TOA ->',
						   title="Cloud Optical Depth Per Layer",
						  plot_width=300, plot_height=300)

	f01.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

	color_bar = ColorBar(color_mapper=color_mapper, ticker=LogTicker(),
					   label_standoff=12, border_line_color=None, location=(0,0))
	f01.add_layout(color_bar, 'left')

	#PLOT G0
	scat01 = np.flip(np.reshape(dat01['g0'].values,(nlayer,nwno)),0)

	xr, yr = scat01.shape
	cols = colfun3(200)[::-1]
	color_mapper = LinearColorMapper(palette=cols, low=0, high=1)


	f01b = figure(x_range=[0, yr], y_range=[0,xr],
						   x_axis_label='Wavenumber Grid', y_axis_label='Pressure Grid, TOA ->',
						   title="Assymetry Parameter",
						  plot_width=300, plot_height=300)

	f01b.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

	color_bar = ColorBar(color_mapper=color_mapper, ticker=BasicTicker(),
					   label_standoff=12, border_line_color=None, location=(0,0))
	f01b.add_layout(color_bar, 'left')
	return column(row(f01a, f01, row(f01b)))

def cloud(full_output):
	"""
	Plotting the cloud input from ``picaso``. 

	The plot itselfs creates maps of the wavelength dependent single scattering albedo 
	and cloud opacity as a function of altitude. 


	Parameters
	----------
	full_output

	Returns
	-------
	A row of two bokeh plots with the single scattering and optical depth map
	"""
	cols = colfun1(200)
	color_mapper = LinearColorMapper(palette=cols, low=0, high=1)

	dat01 = full_output['layer']['cloud']

	#PLOT W0
	scat01 = np.flip(dat01['w0'],0)#[0:10,:]
	xr, yr = scat01.shape
	f01a = figure(x_range=[0, yr], y_range=[0,xr],
						   x_axis_label='Wavenumber Grid', y_axis_label='Pressure Grid, TOA ->',
						   title="Single Scattering Albedo",
						  plot_width=300, plot_height=300)


	f01a.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw = yr)

	color_bar = ColorBar(color_mapper=color_mapper, #ticker=LogTicker(),
					   label_standoff=12, border_line_color=None, location=(0,0))

	f01a.add_layout(color_bar, 'left')


	#PLOT OPD
	scat01 = np.flip(dat01['opd']+1e-60,0)

	xr, yr = scat01.shape
	cols = colfun2(200)[::-1]
	color_mapper = LogColorMapper(palette=cols, low=1e-3, high=10)


	f01 = figure(x_range=[0, yr], y_range=[0,xr],
						   x_axis_label='Wavenumber Grid', y_axis_label='Pressure Grid, TOA ->',
						   title="Cloud Optical Depth Per Layer",
						  plot_width=300, plot_height=300)

	f01.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw = yr)

	color_bar = ColorBar(color_mapper=color_mapper, ticker=LogTicker(),
					   label_standoff=12, border_line_color=None, location=(0,0))
	f01.add_layout(color_bar, 'left')

	#PLOT G0
	scat01 = np.flip(dat01['g0']+1e-60,0)

	xr, yr = scat01.shape
	cols = colfun3(200)[::-1]
	color_mapper = LinearColorMapper(palette=cols, low=0, high=1)


	f01b = figure(x_range=[0, yr], y_range=[0,xr],
						   x_axis_label='Wavenumber Grid', y_axis_label='Pressure Grid, TOA ->',
						   title="Assymetry Parameter",
						  plot_width=300, plot_height=300)

	f01b.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw = yr)

	color_bar = ColorBar(color_mapper=color_mapper, ticker=BasicTicker(),
					   label_standoff=12, border_line_color=None, location=(0,0))
	f01b.add_layout(color_bar, 'left')
	return column(row(f01a, f01, row(f01b)))

def lon_lat_to_cartesian(lon_r, lat_r, R = 1):
	"""
	calculates lon, lat coordinates of a point on a sphere with
	radius R
	"""
	x =  R * np.cos(lat_r) * np.cos(lon_r)
	y = R * np.cos(lat_r) * np.sin(lon_r)
	z = R * np.sin(lat_r)
	return x,y,z

def disco(full_output,wavelength=[0.3]):
	"""
	Plot disco ball with facets. Bokeh is not good with 3D things. So this is in matplotlib

	Parameters
	----------
	full_output : class 
		Full output from picaso

	wavelength : list 
		Where to plot 3d facets. Can input as many wavelengths as wanted. 
		Must be a list, must be in microns. 
	"""

	nrow = int(np.ceil(len(wavelength)/3))
	ncol = int(np.min([3,len(wavelength)])) #at most 3 columns
	fig = plt.figure(figsize=(6*ncol,4*nrow))
	for i,w in zip(range(len(wavelength)),wavelength):
		ax = fig.add_subplot(nrow,ncol,i+1, projection='3d')
		#else:ax = fig.gca(projection='3d')
		wave = 1e4/full_output['wavenumber']
		indw = find_nearest(wave,w)
		#[umg, numt, nwno] this is xint_at_top
		xint_at_top = full_output['albedo_3d'][:,:,indw]

		latitude = full_output['latitude']  #tangle
		longitude = full_output['longitude'] #gangle

		cm = plt.cm.get_cmap('plasma')
		u, v = np.meshgrid(longitude, latitude)
		
		x,y,z = lon_lat_to_cartesian(u, v)

		ax.plot_wireframe(x, y, z, color="gray")

		sc = ax.scatter(x,y,z, c = xint_at_top.T.ravel(),cmap=cm,s=150)

		fig.colorbar(sc)
		ax.set_zlim3d(-1, 1)					# viewrange for z-axis should be [-4,4]
		ax.set_ylim3d(-1, 1)					# viewrange for y-axis should be [-2,2]
		ax.set_xlim3d(-1, 1)
		ax.view_init(0, 0)
		ax.set_title(str(w)+' Microns')
		# Hide grid lines
		ax.grid(False)

		# Hide axes ticks
		ax.set_xticks([])
		ax.set_yticks([])
		ax.set_zticks([])
		plt.axis('off')

	plt.subplots_adjust(wspace=0.3, hspace=0.3)
	plt.show()

#@jit(nopython=True, cache=True)
def find_nearest(array,value):
	#small program to find the nearest neighbor in a matrix
	idx = (np.abs(array-value)).argmin(axis=0)
	return idx
@jit(nopython=True, cache=True)
def numba_cumsum(mat):
	"""Function to compute cumsum along axis=0 to bypass numba not allowing kwargs in 
	cumsum 
	"""
	new_mat = np.zeros(mat.shape)
	for i in range(mat.shape[1]):
		new_mat[:,i] = np.cumsum(mat[:,i])
	return new_mat

def spectrum_hires(wno, alb,legend=None, **kwargs):
	"""Plot formated albedo spectrum

	Parameters
	----------
	wno : float array, list of arrays
		wavenumber 
	alb : float array, list of arrays 
		albedo 
	legend : list of str 
		legends for plotting 
	**kwargs : dict 	
		Any key word argument for hv.opts

	Returns
	-------
	bokeh plot
	"""
	kwargs['plot_height'] = kwargs.get('plot_height',345)
	kwargs['plot_width'] = kwargs.get('plot_width',1000)
	kwargs['y_axis_label'] = kwargs.get('y_axis_label','Albedo')
	kwargs['x_axis_label'] = kwargs.get('x_axis_label','Wavelength [μm]')
	kwargs['y_range'] = kwargs.get('y_range',[0,1.2])
	kwargs['x_range'] = kwargs.get('x_range',[0.3,1])

	points_og = datashade(hv.Curve((1e4/wno, alb)))

	return points_og


	